from dpp.core.plugin import HasherPlugin

class Plugin(HasherPlugin):
    """
    Hashes a string using KECCAK 512.

    Example:

        Input:
            abcdefghijklmnopqrstuvwxyz
            ^°!"§$%&/()=?´`<>| ,.-;:_#+'*~
            0123456789

        Output:
            e7eb5bc85e3d05ccab9863189ce1a34ef2a3feda8ce633b690dd133242a43e90 \
            44bd36c27cb30d66a5eef8b4cb917c609f6983e2c2b8625c0aedb3f87f364172
    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('KECCAK 512', "Thomas Engel", ["_pysha3"], context)

    def run(self, text):
        import _pysha3
        return _pysha3.keccak_512(text.encode('utf-8', errors='surrogateescape')).hexdigest()



from dpp.core.plugin import EncoderPlugin

class Plugin(EncoderPlugin):
    """
    Encodes to hex shell code.

    Example:

        Input:
            abcdefghijklmnopqrstuvwxyz
            ^°!"§$%&/()=?´`<>| ,.-;:_#+'*~
            0123456789

        Output:
            \\x61\\x62\\x63\\x64\\x65\\x66\\x67\\x68\\x69\\x6a\\x6b\\x6c\\x6d\\x6e \\
            \\x6f\\x70\\x71\\x72\\x73\\x74\\x75\\x76\\x77\\x78\\x79\\x7a\\x0a\\x5e \\
            \\xc2\\xb0\\x21\\x22\\xc2\\xa7\\x24\\x25\\x26\\x2f\\x28\\x29\\x3d\\x3f \\
            \\xc2\\xb4\\x60\\x3c\\x3e\\x7c\\x20\\x2c\\x2e\\x2d\\x3b\\x3a\\x5f\\x23 \\
            \\x2b\\x27\\x2a\\x7e\\x0a\\x30\\x31\\x32\\x33\\x34\\x35\\x36\\x37\\x38 \\
            \\x39
    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('HEX (shell)', "Thomas Engel", ["codecs"], context)

    def run(self, text):
        if text:
            import codecs
            output = codecs.encode(text.encode('utf-8', errors='surrogateescape'), 'hex').decode('utf-8', errors='surrogateescape')
            return "\\x" + "\\x".join([i+j for i, j in zip(output[::2], output[1::2])])
        else:
            return ""



from dpp.core.plugin import HasherPlugin

class Plugin(HasherPlugin):
    """
    Hashes a string using KECCAK 224.

    Example:

        Input:
            abcdefghijklmnopqrstuvwxyz
            ^°!"§$%&/()=?´`<>| ,.-;:_#+'*~
            0123456789

        Output:
            dbf87ddd2f01eb7f172b18d94baf83ace62cb71c6ec2b5c82bdf2bab
    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('KECCAK 224', "Thomas Engel", ["_pysha3"], context)

    def run(self, text):
        import _pysha3
        return _pysha3.keccak_224(text.encode('utf-8', errors='surrogateescape')).hexdigest()



# vim: ts=8:sts=8:sw=8:noexpandtab
#
# This file is part of Decoder++
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from PyQt5.QtWidgets import QWidget, QAction


class Shortcut(QAction):
    """ Defines a shortcut with a unique identifier. """

    def __init__(self, id: str, name: str, shortcut_key: str, callback, widget):
        """
        Initializes a shortcut.
        :param id: the unique id of the shortcut (e.g. "next_frame_focus").
        :param name: the name of the shortcut which is displayed to the user (e.g. "Focus next frame").
        :param shortcut_key: the shortcut key which triggers an action (e.g. "Alt+Down").
        :param callback: the callback which should be triggered when the shortcut key is pressed.
        :param widget: the widget on which the shortcut is bound to.
        """
        super(__class__, self).__init__(name, widget)
        self._id = id
        self._name = name
        self.setShortcut(shortcut_key)
        self.triggered.connect(callback)

    def id(self) -> str:
        """ Returns the unique id of the shortcut (e.g. "next_frame_focus"). """
        return self._id

    def name(self, remove_anchors=False) -> str:
        """ Returns the name of the shortcut (e.g. "Focus next frame").
        :param remove_anchors: if True removes every occurences of the ampersand-symbol (&) within the name.
        """
        if remove_anchors:
            return self._name.replace("&", "")
        return self._name

    def setName(self, name: str):
        """ Sets the name of the shortcut which is displayed to the user (e.g. "Focus next frame"). """
        self._name = name

    def key(self) -> str:
        """ Returns the shortcut key (e.g. "Alt+Down") which triggers an action. """
        return self.shortcut().toString()

    def setKey(self, key: str):
        """ Sets the shortcut key (e.g. "Alt+Down") which triggers an action. """
        self.setShortcut(key)


class NullShortcut(Shortcut):
    """ Defines an empty shortcut which is used in case of an error (e.g. when no shortcut was defined). """

    def __init__(self):
        self._widget = QWidget()
        # id, name, shortcut_key, callback, widget
        super(__class__, self).__init__("", "", "", lambda: None, self._widget)



# vim: ts=8:sts=8:sw=8:noexpandtab
#
# This file is part of Decoder++
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from PyQt5.QtCore import QSettings, QSize


class Config(QSettings):
    """ The main settings of the application. """

    def __init__(self):
        super().__init__('net.bytebutcher', 'decoder++')

    def isDebugModeEnabled(self) -> bool:
        """ Returns whether debug mode is enabled (True = enabled, False = disabled). """
        return self.value('debug', "False") == "True"

    def setDebugMode(self, status: bool):
        """ Enables/Disables debug mode. """
        # BUG: Storing boolean True/False does not work (see https://github.com/mfitzp/pyqtconfig/issues/18)
        # FIX: String string "True"/"False" instead.
        self.setValue('debug', "True" if status else "False")

    def getSize(self) -> QSize:
        """
        Returns the saved window size of the application.
        When no size was specified a default size will be returned.
        """
        return self.value('size', QSize(640, 440))

    def getPosition(self) -> int:
        """ Returns the saved position of the window."""
        return self.value('position')

    def setSize(self, size: QSize):
        """ Stores the size of the window. """
        self.setValue('size', size)

    def setPosition(self, position: int):
        """ Stores the position of the window. """
        self.setValue('position', position)

    def setShortcutKey(self, id: str, shortcut_key: str):
        """ Stores a shortcut key. """
        self.setValue('shortcut.{}'.format(id), shortcut_key)

    def getShortcutKey(self, id: str):
        """ Returns a saved shortcut key or None when no shortcut were defined for the specified id. """
        return self.value('shortcut.{}'.format(id))

    def setPluginStatus(self, id: str, status: bool):
        """ Sets the status of the specified plugin to enabled/disabled. """
        self.setValue('plugin.{}'.format(id.lower()), str(status))

    def getPluginStatus(self, id: str) -> bool:
        """ Returns whether the plugin is enabled/disabled. When no status was stored True will be returned. """
        status = self.value('plugin.{}'.format(id.lower()))
        return status == "True" or status is None




# vim: ts=8:sts=8:sw=8:noexpandtab
#
# This file is part of Decoder++
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import QFrame

from dpp.core import Context


class Frame(QFrame):
    """
    Encapsulates the basic frame behaviour.
    """

    frameChanged = pyqtSignal(str)

    def __init__(self, parent, context: Context, frame_id: str):
        super(__class__, self).__init__(parent)
        self._frame_id = frame_id
        self._context = context
        self._listener = self._context.listener()
        self._logger = context.logger()

    def id(self) -> str:
        """ Returns the individual identifier of the frame. """
        return self._frame_id

    def isConfigurable(self) -> bool:
        """
        Returns whether the frame is configurable (always False).
        This method needs to be overridden if other behaviour is required.
        """
        return False



# vim: ts=8:sts=8:sw=8:noexpandtab
#
# This file is part of Decoder++
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from PyQt5.QtWidgets import QWidget, QFrame, QHBoxLayout, QStackedWidget


class DynaFrame(QFrame):
    """ The DynaFrame class wraps one custom widget which can be replaced with another during runtime. """

    def __init__(self, parent, widget: QWidget=None):
        super(__class__, self).__init__(parent)
        layout = QHBoxLayout()
        self._stk_widget = QStackedWidget()
        if widget is not None:
            self.setWidget(widget)
        layout.addWidget(self._stk_widget)
        self.setLayout(layout)

    def setWidget(self, widget: QWidget):
        """ Replaces the currently present widget (if any) with the specified one. """
        self._clear_stack_widget()
        self._stk_widget.addWidget(widget)

    def getWidget(self) -> QWidget:
        """ Returns the widget wrapped in the frame. """
        return self._stk_widget.widget(0)

    def _clear_stack_widget(self):
        """ Clears all widgets which were added to the stack. """
        count = self._stk_widget.count()
        assert 0 <= count <= 1
        for index in range(0, count):
            widget = self._stk_widget.widget(index)
            self._stk_widget.removeWidget(widget)



# vim: ts=8:sts=8:sw=8:noexpandtab
#
# This file is part of Decoder++
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from typing import List

import qtawesome
from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import QLabel, QTabWidget, QToolButton, QMenu, QFrame, QHBoxLayout

from dpp.core import Context
from dpp.ui import CodecTab, TabBar
from dpp.ui.builder.action_builder import ActionBuilder
from dpp.ui.widget.search_field import SearchField
from dpp.ui.widget.separater_widget import VSep


class MainWindowTabsWidget(QTabWidget):

    # tabAdded(index, name)
    tabAdded = pyqtSignal('PyQt_PyObject', 'PyQt_PyObject')
    # tabClosed(index, name)
    tabClosed = pyqtSignal('PyQt_PyObject', 'PyQt_PyObject')
    # tabRenamed(index, old_name, new_name)
    tabRenamed = pyqtSignal('PyQt_PyObject', 'PyQt_PyObject', 'PyQt_PyObject')
    # tabMovedToPrevious(old_index, new_index, name)
    tabMovedToPrevious = pyqtSignal('PyQt_PyObject', 'PyQt_PyObject', 'PyQt_PyObject')
    # tabMovedToNext(old_index, new_index, name)
    tabMovedToNext = pyqtSignal('PyQt_PyObject', 'PyQt_PyObject', 'PyQt_PyObject')

    def __init__(self, context: Context, plugins, parent=None):
        super(__class__, self).__init__(parent)
        self._parent = parent
        self._context = context
        self._current_tab_number = 1
        self._init_listener(context)
        self._plugins = plugins
        bar = TabBar()
        bar.customContextMenuRequested.connect(self._show_context_menu)
        bar.tabRenamed.connect(self.tabRenamed.emit)
        self.setTabBar(bar)
        self.setTabsClosable(True)
        self.tabCloseRequested.connect(self.closeTab)
        tab_new_button = QToolButton()
        tab_new_button.setToolTip("New Tab")
        tab_new_button.clicked.connect(lambda: self.newTab())
        tab_new_button.setIcon(qtawesome.icon("fa.plus"))
        self.addTab(QLabel("Add new Tab"), "")
        self.setTabEnabled(0, False)
        self.tabBar().setTabButton(0, TabBar.RightSide, tab_new_button)

    def _init_search_field(self):
        search_field_frame = QFrame()
        search_field_layout = QHBoxLayout()
        search_field_layout.addWidget(VSep(self))

        self._search_field = SearchField(self)
        self._search_field.setIcon(qtawesome.icon("fa.search"))
        self._search_field.setPlaceholderText("Search codec")
        self._search_field.textChanged.connect(lambda text: {})
        search_field_layout.addWidget(self._search_field)

        self._search_field_button = QToolButton()
        self._search_field_button.setIcon(qtawesome.icon("fa.arrow-circle-right"))
        self._search_field_button.setContentsMargins(0, 0, 0, 0)
        search_field_layout.addWidget(self._search_field_button)

        search_field_frame.setLayout(search_field_layout)
        search_field_frame.setContentsMargins(5, 0, 5, 5) # left, top, right, bottom
        search_field_frame.layout().setContentsMargins(0, 0, 0, 0)
        return search_field_frame

    def _init_listener(self, context):
        self._listener = context.listener()
        self._listener.newTabRequested.connect(self.newTab)

    def _show_context_menu(self, point):
        if not point:
            return

        index = self.tabBar().tabAt(point)
        if index == self.count() - 1:
            # Do not show any context-menu for "Add new tab"-Tab.
            return

        menu = QMenu(self)
        menu.addAction(ActionBuilder(self)
                       .name("New Tab")
                       .callback(lambda checked: self.newTab()).build())
        menu.addSeparator()
        menu.addAction(ActionBuilder(self)
                       .name("Rename Tab")
                       .callback(lambda checked: self.tabBar().renameTab()).build())
        menu.addAction(ActionBuilder(self)
                       .name("Duplicate Tab")
                       .callback(lambda checked: self.duplicateTab()).build())
        menu.addSeparator()
        menu.addAction(ActionBuilder(self)
                       .name("Close Tab")
                       .callback(lambda checked: self.closeTab()).build())
        menu.addAction(ActionBuilder(self)
                       .name("Close Other Tabs")
                       .enabled(self.tabCount() > 1)
                       .callback(lambda checked: self.closeOtherTabs()).build())
        menu.exec(self.tabBar().mapToGlobal(point))

    def newTab(self, input: str=None, title :str=None) -> (int, CodecTab):
        """
        Opens a new tab and writes input into first codec-frame.
        :param input: The input which should be placed into the first codec-frame.
        :param title: The title of the tab.
        """
        codec_tab = CodecTab(self, self._context, self._plugins)
        name = "Tab {}".format(self._current_tab_number) if not title else title
        self._current_tab_number += 1
        self.insertTab(self.count() - 1, codec_tab, name)
        index = self.count() - 2
        self.setCurrentIndex(index)
        self.tabAdded.emit(index, name)
        codec_tab.frames().getFocusedFrame().setInputText(input)
        # BUG: Input-text of newly added codec-tab is not focused correctly.
        # FIX: Refocus input-text.
        codec_tab.frames().getFocusedFrame().focusInputText()
        return index, codec_tab

    def renameTab(self, index, title):
        """ Renames the title of a tab.
        Note: If you require the user to dynamically enter a value checkout the renameTab method in tab_bar.py
        :param index: the index of the tab to be renamed.
        """
        self.tabBar().setTabText(index, title)
        return self.tab(index)

    def duplicateTab(self, src_index=None):
        """
        Duplicates the content of a tab.
        :param src_index: the index of the tab to be duplicated.
        """
        def does_tab_title_already_exist(title):
            for index in range(self.count()):
                if self.tabBar().tabText(index) == title:
                    return True
            return False

        def get_duplicate_tab_title(index):
            count = 1
            while does_tab_title_already_exist(self.tabBar().tabText(index) + " (Copy " + str(count) + ")"):
                count += 1
            return self.tabBar().tabText(index) + " (Copy " + str(count) + ")"

        if src_index is None:
            src_index = self.currentIndex()

        src_tab = self.tab(src_index)
        index, tab = self.newTab(title=get_duplicate_tab_title(src_index))
        frame_index = 0
        for src_frame in src_tab.frames().getFrames():
            status_type, status_message = src_frame.status()
            if frame_index == 0:
                # New tabs already contain one empty frame
                frame = tab.frames().getFrames()[0]
                frame.setInputText(src_frame.getInputText())
                frame.setStatus(status_type, status_message)
            else:
                tab.frames().newFrame(src_frame.getInputText(),
                                              src_frame.title(),
                                              frame_index,
                                              status_type,
                                              status_message)
            tab.frames().getFrameByIndex(frame_index).fromDict(src_frame.toDict())
            frame_index = frame_index + 1

    def selectTab(self, index):
        if index < 0 or index > self.count() - 2:
            return
        self.setCurrentIndex(index)

    def closeTab(self, index=None):
        if index is None:
            index = self.currentIndex()
        name = self.tabText(index)
        self.removeTab(index)
        if self.count() <= 1:
            self.newTab()
        else:
            self.setCurrentIndex(index - 1)
        self.tabClosed.emit(index, name)

    def closeOtherTabs(self, current_index=None):
        if current_index is None:
            current_index = self.currentIndex()

        # Close every tab ahead of the current one, starting at the end.
        for index in reversed(range(current_index + 1, self.tabCount())):
            self.closeTab(index)

        # Close every tab below the current one, starting at the beginning.
        for index in range(0, current_index):
            self.closeTab(0)

    def nextTab(self):
        old_index = self.currentIndex()
        name = self.tabText(old_index)
        new_index = (old_index + 1) % (self.count() - 1)
        self.setCurrentIndex(new_index)
        self.tabMovedToNext.emit(old_index, new_index, name)

    def previousTab(self):
        old_index = self.currentIndex()
        name = self.tabText(old_index)
        new_index = (old_index - 1) % (self.count() - 1)
        self.setCurrentIndex(new_index)
        self.tabMovedToPrevious.emit(old_index, new_index, name)

    def tab(self, index) -> CodecTab:
        return self.widget(index)

    def tabCount(self):
        # do not count the "Add new Tab"-Tab.
        return self.count() - 1

    def focusCodecSearch(self):
        self._search_field.setFocus()

    def toDict(self) -> List[dict]:
        return [ {
            "name": self.tabBar().tabText(tabIndex),
            "frames": self.tab(tabIndex).toDict()
        } for tabIndex in range(0, self.tabCount()) ]



from dpp.core.plugin import DecoderPlugin

class Plugin(DecoderPlugin):
    """
    Decodes an octal string to an integer.

    Example:

        Input:
            726746425

        Output:
            123456789
    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('OCT (int)', "Thomas Engel", [], context)

    def run(self, text):
        return self._run_lines(text, lambda text_part: str(int(text_part, 8)))



from dpp.core.plugin import DecoderPlugin

class Plugin(DecoderPlugin):
    """
    Decodes a HTML string.

    Example:

        Input:
            abcdefghijklmnopqrstuvwxyz
            ^°!&quot;§$%&amp;/()=?´`&lt;&gt;| ,.-;:_#+&#x27;*~
            0123456789

        Output:
            abcdefghijklmnopqrstuvwxyz
            ^°!"§$%&/()=?´`<>| ,.-;:_#+'*~
            0123456789
    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('HTML', "Thomas Engel", ["html"], context)

    def run(self, text):
        import html
        return html.unescape(text)

    def can_decode_input(self, input):
        """
        Checks whether input can be decoded. When decoded input does not match the initial input some decoding must
        have happened so we return True.
        """
        if input:
            try:
                return self.run(input) != input
            except:
                return False
        return False



import os
import unittest

test_path = os.path.dirname(os.path.realpath(__file__))


def load_tests(loader, filter):
    """
    Load all test cases and return a unittest.TestSuite object.
    """
    suite = unittest.TestSuite()
    for r, d, f in os.walk(test_path):
        module_path = os.path.relpath(r, os.path.dirname(test_path)).replace(os.path.sep, ".")
        for file in f:
            filename, ext = os.path.splitext(file)
            if ext == ".py" and not file.startswith("_"):
                if not filter or module_path + "." + filename in filter:
                    print("[ADD ] " + module_path + "." + filename)
                    test = loader.loadTestsFromName(module_path + "." + filename)
                    suite.addTest(test)
                else:
                    print("[SKIP] " + module_path + "." + filename)
    return suite



from dpp.core.plugin import DecoderPlugin

class Plugin(DecoderPlugin):
    """
    Decodes an URL.

    Example:

        Input:
            abcdefghijklmnopqrstuvwxyz \\
            %0A%5E%C2%B0%21%22%C2%A7%24%25%26/%28%29%3D%3F%C2%B4%60%3C%3E%7C%20%2C.-%3B%3A_%23%2B%27%2A%7E%0A \\
            0123456789

        Output:
            abcdefghijklmnopqrstuvwxyz
            ^°!"§$%&/()=?´`<>| ,.-;:_#+'*~
            0123456789
    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('URL', "Thomas Engel", ["urllib"], context)

    def run(self, text):
        import urllib.parse
        return urllib.parse.unquote(text)

    def can_decode_input(self, input):
        """
        Checks whether input can be decoded. When the input contains a plus sign we return False since it is more likely
        that our input is URL+ encoded instead.  When decoded input does not match the initial input some decoding must
        have happened so we return True.
        """
        if input and "+" not in input:
            try:
                return self.run(input) != input
            except:
                return False
        return False



from dpp.core.plugin import DecoderPlugin

class Plugin(DecoderPlugin):
    """
    Decodes a Base16 string.

    Example:

        Input:
            6162636465666768696A6B6C6D6E6F70717273747576777 \\
            8797A0A5EC2B02122C2A72425262F28293D3FC2B4603C3E \\
            7C202C2E2D3B3A5F232B272A7E0A30313233343536373839

        Output:
            abcdefghijklmnopqrstuvwxyz
            ^°!"§$%&/()=?´`<>| ,.-;:_#+'*~
            0123456789
    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('BASE16', "Thomas Engel", ["base64"], context)

    def run(self, text):
        import base64
        return base64.b16decode(text.encode('utf-8', errors='surrogateescape')).decode('utf-8', errors='surrogateescape')



from dpp.core.plugin import HasherPlugin

class Plugin(HasherPlugin):
    """
    Hashes a string using SHA3 224.

    Example:

        Input:
            abcdefghijklmnopqrstuvwxyz
            ^°!"§$%&/()=?´`<>| ,.-;:_#+'*~
            0123456789

        Output:
            0049b8b6c811930a8873b7b14bd5191e56931661626e89248ee476e9
    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('SHA3 224', "Thomas Engel", ["_pysha3"], context)

    def run(self, text):
        import _pysha3
        return _pysha3.sha3_224(text.encode('utf-8', errors='surrogateescape')).hexdigest()



from dpp.core.plugin import EncoderPlugin

class Plugin(EncoderPlugin):
    """
    Encodes an integer to a binary string.

    Example:

        Input:
            123456789

        Output:
            111010110111100110100010101
    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('BIN (int)', "Thomas Engel", [], context)

    def run(self, text):
        return self._run_lines(text, lambda text_part: "{0:b}".format(int(text_part)))



from dpp.core.plugin import EncoderPlugin

class Plugin(EncoderPlugin):
    """
    Encodes a string to a binary string.

    Example:

        Input:
            abcdefghijklmnopqrstuvwxyz
            ^°!"§$%&/()=?´`<>| ,.-;:_#+'*~
            0123456789

        Output:
            1100001 1100010 1100011 1100100 1100101 \\
            1100110 1100111 1101000 1101001 1101010 \\
            ...
            110100 110101 110110 110111 111000 111001
    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('BIN (str)', "Thomas Engel", ["codecs"], context)

    def run(self, text):
        bits = bin(int.from_bytes(text.encode('utf-8', 'surrogateescape'), 'big'))[2:]
        return ' '.join(self._chunk_string(bits.zfill(8 * ((len(bits) + 7) // 8)), 8))

    def _chunk_string(self, string, length):
        return [string[0 + i:length + i] for i in range(0, len(string), length)]



from dpp.core.plugin import EncoderPlugin

class Plugin(EncoderPlugin):
    """
    Encodes a text using HTTP64.

    HTTP64 is a BASE64 derivation which remaps following characters:

     Location  Base64  Http64
     --------  ------  ------
       62        +       -
       63        /       _
      (pad)      =       $

    Example:

        Input:
            abcdefghijklmnopqrstuvwxyz
            ^°!"§$%&/()=?´`<>| ,.-;:_#+'*~
            0123456789

        Output:
            YWJjZGVmZ2hpamtsbW5vcHFyc3R1dnd4eXoKXsKwISLCpyQlJi8oKT0_wrRgPD58ICwuLTs6XyMrJyp-CjAxMjM0NTY3ODk$
    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('HTTP64', "Thomas Engel", ["base64"], context)

    def run(self, text):
        import base64
        encoded = base64.b64encode(text.encode('utf-8', errors="surrogateescape")).decode('utf-8', errors="surrogateescape")
        encoded = encoded.replace('+', '-')
        encoded = encoded.replace('/', '_')
        encoded = encoded.replace('=', '$')
        return encoded




from dpp.core.plugin import HasherPlugin

class Plugin(HasherPlugin):
    """
    Hashes a string using SHA3 512.

    Example:

        Input:
            abcdefghijklmnopqrstuvwxyz
            ^°!"§$%&/()=?´`<>| ,.-;:_#+'*~
            0123456789

        Output:
            82ca87f576cadb05d4c911f36c98ed2735f45cad359d6ef5f6d544f5a3210e3e \
            cf080be15e539e23c15e2eb23054677d8a015ee56be2d9673c9f187d290906ed
    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('SHA3 512', "Thomas Engel", ["_pysha3"], context)

    def run(self, text):
        import _pysha3
        return _pysha3.sha3_512(text.encode('utf-8', errors='surrogateescape')).hexdigest()



from dpp.core.plugin import HasherPlugin

class Plugin(HasherPlugin):
    """
    Hashes a string using SHA512.

    Example:

        Input:
            abcdefghijklmnopqrstuvwxyz
            ^°!"§$%&/()=?´`<>| ,.-;:_#+'*~
            0123456789

        Output:
            58cd501bee1ece7411a23b2e86bfb795b136f2aae5d8f94a59c551622d9a0d7e \\
            293a04a8584244eadf1f9bedd34cbcb7e99f7bdedaac56591f88bb282c5146cd

    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('SHA512', "Thomas Engel", ["hashlib"], context)

    def run(self, text):
        import hashlib
        return hashlib.sha512(text.encode('utf-8', errors='surrogateescape')).hexdigest()



# vim: ts=8:sts=8:sw=8:noexpandtab
#
# This file is part of Decoder++
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from PyQt5.QtCore import QObject, pyqtSignal
from PyQt5.QtNetwork import QLocalServer, QLocalSocket

class InstanceHandler(QObject):
    """ Makes sure that only one instance of this application can be run. """

    received = pyqtSignal(str)

    def __init__(self, parent, name):
        super(__class__, self).__init__(parent)
        self._parent = parent
        self._name = name
        self._timeout = 1000
        self._socket = QLocalSocket(self)
        self._socket.connectToServer(self._name)
        self._is_already_running = self._socket.waitForConnected(self._timeout)
        if not self.isAlreadyRunning():
            self._server = QLocalServer(self)
            self._server.newConnection.connect(self._receive_data)
            self._server.removeServer(self._name)
            self._server.listen(self._name)

    def _send_data(self, data=None):
        """ Sends model to an server-application. """
        if not data:
            data = ""
        self._socket.write(data.encode())
        self._socket.waitForBytesWritten(self._timeout)
        self._socket.disconnectFromServer()

    def _receive_data(self):
        """ Receives model from an client-application. """
        socket = self._server.nextPendingConnection()
        if socket.waitForReadyRead(self._timeout):
            # Emit transmitted model.
            self.received.emit(socket.readAll().data().decode("utf-8", "surrogateescape"))
        else:
            # When no model was transmitted just emit empty string.
            self.received.emit("")

    def newTab(self, input):
        """ Opens a new tab in an already running instance. """
        self._send_data(input)

    def isAlreadyRunning(self) -> bool:
        """ Returns True when an instance of the application is already running, otherwise False. """
        return self._is_already_running



# vim: ts=8:sts=8:sw=8:noexpandtab
#
# This file is part of Decoder++
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import qtawesome
from PyQt5.QtCore import QRect, pyqtSignal, QTimer, Qt
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QFrame, QHBoxLayout, QLabel

from dpp.ui import IconLabel
from dpp.ui.widget.clickable_label import ClickableLabel


class MessageWidget(QFrame):
    """ Widget which shows messages. """

    messageReceived = pyqtSignal('PyQt_PyObject', 'PyQt_PyObject')
    messageClicked = pyqtSignal()
    infoClicked = pyqtSignal()
    errorClicked = pyqtSignal()

    ICON_READY = "ICON_READY"
    ICON_INFO = "ICON_INFO"
    ICON_ERROR = "ICON_ERROR"

    class CountWidget(QFrame):
        """ A widget with an icon and a counter. """

        icon_clicked = pyqtSignal()

        def __init__(self, icon: QIcon, count: int=0, parent=None):
            super(__class__, self).__init__(parent)
            self._icon = IconLabel(parent, icon)
            self._icon.setHoverEffect(True)
            self._icon.clicked.connect(lambda: self.icon_clicked.emit())
            self._label = QLabel(self)
            self.setCount(count)
            layout = QHBoxLayout()
            layout.addWidget(self._icon)
            layout.addWidget(self._label)
            layout.setContentsMargins(0, 0, 0, 0)
            self.setLayout(layout)

        def setCount(self, count: int):
            """ Sets the counter to the specified value. """
            self._label.setText(str(count))

        def getCount(self) -> int:
            """ Returns the counter. """
            return int(self._label.text())

        def incrementCount(self):
            """ Increments the counter. """
            self._label.setText(str(self.getCount() + 1))

        def decrementCount(self):
            """ Decrements the counter. """
            self._label.setText(str(self.getCount() - 1))

        def resetCount(self):
            """ Resets the counter to zero. """
            self.setCount(0)

    def __init__(self, parent=None):
        super(__class__, self).__init__(parent)
        layout = QHBoxLayout()

        self._icons = dict()
        self._icons[self.ICON_READY] = qtawesome.icon("fa.check")
        self._icons[self.ICON_INFO] = qtawesome.icon("fa.info")
        self._icons[self.ICON_ERROR] = qtawesome.icon("fa.exclamation")

        self._log_info_count_widget = MessageWidget.CountWidget(qtawesome.icon("fa.info-circle"), 0, self)
        self._log_info_count_widget.icon_clicked.connect(lambda: self.infoClicked.emit())
        self._log_info_count_widget.mouseDoubleClickEvent = lambda e: self.infoClicked.emit()
        self._log_error_count_widget = MessageWidget.CountWidget(qtawesome.icon("fa.exclamation-triangle"), 0, self)
        self._log_error_count_widget.icon_clicked.connect(lambda: self.errorClicked.emit())
        self._log_error_count_widget.mouseDoubleClickEvent = lambda e: self.errorClicked.emit()

        layout.addWidget(self._log_info_count_widget)
        layout.addWidget(self._log_error_count_widget)

        line = QFrame(self)
        line.setGeometry(QRect(320, 150, 118, 3))
        line.setFrameShape(QFrame.VLine)
        line.setFrameShadow(QFrame.Sunken)
        layout.addWidget(line, 0, Qt.AlignLeft)

        self._log_message_icon_label = IconLabel(self, self._get_icon(self.ICON_READY))
        self._log_message_icon_label.setHoverEffect(True)
        self._log_message_icon_label.clicked.connect(lambda evt: self.messageClicked.emit())
        self._log_message_icon_label.setHoverEffect(True)
        self._log_message_text_label = ClickableLabel("Ready.")
        self._log_message_text_label.doubleClicked.connect(lambda e: self.messageClicked.emit())
        layout.addWidget(self._log_message_icon_label)
        layout.addWidget(self._log_message_text_label)

        self.setLayout(layout)

    def _get_icon(self, type: str):
        """
        Returns an icon of the specified type.
        :param type: the type of the icon (e.g. READY, INFO, ERROR).
        """
        return self._icons[type]

    def _log_message(self, type: str, message: str):
        """
        Emits a message event.
        :param type: the type of the message (e.g. info, error, debug).
        :param message: the message.
        """
        self.messageReceived.emit(type, message)


    def showDebug(self, message: str, log_only: bool=False):
        """ Adds debug message to the log. """
        self._log_message("DEBUG", message)

    def showInfo(self, message: str, log_only: bool=False):
        """ Shows info message in the statusbar. Adds info message to the log. Increments info count. """
        self._log_message("INFO", message)
        self._log_info_count_widget.incrementCount()
        if not log_only:
            self.showMessage(message, self.ICON_INFO)

    def showError(self, message: str, log_only: bool=False):
        """ Shows error message in the statusbar. Adds error message to the log. Increments error count. """
        self._log_message("ERROR", message)
        self._log_error_count_widget.incrementCount()
        if not log_only:
            self.showMessage(message, self.ICON_ERROR)

    def showMessage(self, message: str, icon_type: str):
        """
        Displays a message (shortened to 100 characters) for 5 seconds in the statusbar.
        :param message: the message to display.
        :param icon_type: the type of the icon (e.g. INFO, ERROR).
        """
        if message:
            def _show_ready():
                self._log_message_icon_label.setIcon(self._get_icon(self.ICON_READY))
                self._log_message_text_label.setText("Ready.")

            message = self._shorten_message(message, 100)
            self._log_message_icon_label.setIcon(self._get_icon(icon_type))
            self._log_message_text_label.setText(message)
            QTimer.singleShot(5000, _show_ready)

    def _shorten_message(self, message: str, max_length: int):
        """
        Removes new-lines and shortens message to the specified max_length.
        :param message: the message to shorten.
        :param max_length: the max length of the message.
        :return: the shortened message.
        """
        if not message:
            return ""
        merged_lines = " ".join(message.splitlines())
        if len(merged_lines) > (max_length - 3):
            return merged_lines[:(max_length - 3)] + "..."
        else:
            return merged_lines

    def resetCount(self):
        """ Resets the info- and error-log-counters to zero. """
        self._log_info_count_widget.resetCount()
        self._log_error_count_widget.resetCount()



from dpp.core.plugin import ScriptPlugin


class Plugin(ScriptPlugin):
    """
    Transforms valid JSON- into XML-structure.
    """

    def __init__(self, context: 'core.context.Context'):
        # Name, Author, Dependencies
        super().__init__('JS to XML', "Thomas Engel", ["json2xml"], context)

    def run(self, text):
        from json2xml import json2xml, readfromstring
        return json2xml.Json2xml(readfromstring(text)).to_xml()



from dpp.core.plugin import ScriptPlugin


class Plugin(ScriptPlugin):
    """
    Transforms a hex string from little to big endian vice versa.

    Example:

        Input:
            0002000A

        Output:
            0A000200
    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('Little/Big Endian', "Thomas Engel", [], context)

    def title(self):
        return "Little/Big Endian Transform"

    def run(self, text):
        return ''.join(list(reversed(self._chunk_string(text, 2))))

    def _chunk_string(self, string, length):
        return [string[0 + i:length + i] for i in range(0, len(string), length)]


# vim: ts=8:sts=8:sw=8:noexpandtab
#
# This file is part of Decoder++
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from PyQt5.QtCore import QSortFilterProxyModel, Qt
from PyQt5.QtWidgets import QCompleter, QComboBox


class ComboBox(QComboBox):
    def __init__(self,  parent = None):
        super(ComboBox, self).__init__(parent)

        self.setFocusPolicy(Qt.StrongFocus)
        self.setEditable(True)
        self.completer = QCompleter(self)

        # always show all completions
        self.completer.setCompletionMode(QCompleter.UnfilteredPopupCompletion)
        self.pFilterModel = QSortFilterProxyModel(self)
        self.pFilterModel.setFilterCaseSensitivity(Qt.CaseInsensitive)
        self.completer.setPopup(self.view())
        self.setCompleter(self.completer)
        self.lineEdit().textEdited.connect(self.pFilterModel.setFilterFixedString)
        self.completer.activated.connect(self.setTextIfCompleterIsClicked)

    def setModel(self, model):
        super(ComboBox, self).setModel(model)
        self.pFilterModel.setSourceModel(model)
        self.completer.setModel(self.pFilterModel)

    def setModelColumn(self, column):
        self.completer.setCompletionColumn(column)
        self.pFilterModel.setFilterKeyColumn(column)
        super(ComboBox, self).setModelColumn(column)

    def view(self):
        return self.completer.popup()

    def index(self):
        return self.currentIndex()

    def setTextIfCompleterIsClicked(self, text):
      if text:
        index = self.findText(text)
        self.setCurrentIndex(index)


# vim: ts=8:sts=8:sw=8:noexpandtab
#
# This file is part of Decoder++
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from PyQt5.QtCore import QSortFilterProxyModel, pyqtSignal, Qt
from PyQt5.QtGui import QStandardItemModel, QStandardItem, QKeyEvent
from PyQt5.QtWidgets import QTableView
from qtpy import QtWidgets

from dpp.ui.dialog.keyboard_shortcut_dialog import KeyboardShortcutDialog


class KeyboardShortcutTable(QTableView):

    changed = pyqtSignal('PyQt_PyObject', 'PyQt_PyObject')
    keyPressed = pyqtSignal(str)

    def __init__(self, parent, context):
        super(KeyboardShortcutTable, self).__init__(parent)
        self._context = context
        self._init_model(context.getShortcuts())
        self._init_headers()
        self._init_proxy_model()
        self.clicked.connect(self.clickEvent)

    def _init_headers(self):
        header = self.horizontalHeader()
        header.setSectionResizeMode(0, QtWidgets.QHeaderView.ResizeToContents)
        header.setSectionResizeMode(1, QtWidgets.QHeaderView.ResizeToContents)
        header.setSectionResizeMode(2, QtWidgets.QHeaderView.Stretch)
        self.verticalHeader().hide()

    def _init_model(self, keyboard_shortcuts):
        model = QStandardItemModel(len(keyboard_shortcuts), 3)
        model.setHorizontalHeaderLabels(["Id", "Name", "Keyboard Shortcut"])
        for index, keyboard_shortcut in enumerate(keyboard_shortcuts):
            model.setItem(index, 0, QStandardItem(keyboard_shortcut.id()))
            model.setItem(index, 1, QStandardItem(keyboard_shortcut.name(remove_anchors=True)))
            model.item(index, 1).setFlags(model.item(index, 1).flags() ^ Qt.ItemIsEditable)
            model.setItem(index, 2, QStandardItem(keyboard_shortcut.key()))
            model.item(index, 2).setFlags(model.item(index, 2).flags() ^ Qt.ItemIsEditable)
        self.setModel(model)
        self.setColumnHidden(0, True)

    def _init_proxy_model(self):
        filter_proxy_model = QSortFilterProxyModel()
        filter_proxy_model.setSourceModel(self.model())
        filter_proxy_model.setFilterKeyColumn(1)
        filter_proxy_model.setFilterCaseSensitivity(Qt.CaseInsensitive)
        self.setModel(filter_proxy_model)

    def _edit_keyboard_shortcut(self, item):
        keyboard_shortcut_name = self.model().sourceModel().item(item.row(), 1).text()
        keyboard_shortcut_dialog = KeyboardShortcutDialog(self, self._context, keyboard_shortcut_name)
        keyboard_shortcut_dialog.exec_()
        if keyboard_shortcut_dialog.shouldBeReset():
            self._update_keyboard_shortcut(item, "")
        elif keyboard_shortcut_dialog.keyboardShortcut():
            self._update_keyboard_shortcut(item, keyboard_shortcut_dialog.keyboardShortcut())

    def _update_keyboard_shortcut(self, item, keyboard_shortcut):
        id = self.model().sourceModel().item(item.row(), 0).text()
        self.model().sourceModel().item(item.row(), 2).setText(keyboard_shortcut)
        self.changed.emit(id, keyboard_shortcut)

    def _get_source_model_item(self, item):
        """
        Due to letting users search for keyboard shortcuts we actually have two models:
        1. the source model containing all keyboard shortcuts
        2. the proxy model containing the keyboard shortcuts visible to the user due to his search term
        When an item is clicked Qt returns the item associated with the proxy model which we are likely not interested
        in. This method is a shorthand for mapping the item of the proxy model to the item of the source model.
        """
        return self.model().sourceModel().item(self.model().mapToSource(item).row())

    def clickEvent(self, item):
        self._edit_keyboard_shortcut(self._get_source_model_item(item))

    def keyReleaseEvent(self, event: QKeyEvent):
        src_item = self._get_source_model_item(self.currentIndex())
        if event.key() == Qt.Key_Return:
            self._edit_keyboard_shortcut(src_item)
        elif event.key() == Qt.Key_Backspace:
            self._update_keyboard_shortcut(src_item, "")
        else:
            try:
                # We only allow alpha numeric keys as shortcut like ctrl + "A" but not ctrl + "!".
                if chr(event.key()).isalnum():
                    self.keyPressed.emit(chr(event.key()))
            except:
                # Ignore exceptions which occur when the key is not a char.
                pass



# vim: ts=8:sts=8:sw=8:noexpandtab
#
# This file is part of Decoder++
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import qtawesome

from dpp.ui import IconLabel


class VSep(IconLabel):
    """ Widget which draws a vertical separator. """

    def __init__(self, parent=None):
        super(VSep, self).__init__(parent, qtawesome.icon("fa.ellipsis-v"))
        self.setDisabled(True) # just for the effect


class HSep(IconLabel):
    """ Widget which draws a horizontal separator. """

    def __init__(self, parent=None):
        super(VSep, self).__init__(parent, qtawesome.icon("fa.ellipsis-h"))
        self.setDisabled(True) # just for the effect



# vim: ts=8:sts=8:sw=8:noexpandtab
#
# This file is part of Decoder++
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import uuid
from typing import List

from PyQt5.QtCore import Qt, pyqtSignal
from PyQt5.QtWidgets import QScrollArea, QWidget, QFrame, QVBoxLayout, QSplitter, QSizePolicy

from dpp.core import Context
from dpp.core.plugin import Plugins, PluginBuilder
from dpp.ui.codec_frames import CodecFrames
from dpp.ui.widget.spacer import VSpacer
from dpp.ui.widget.status_widget import StatusWidget


class CodecTab(QScrollArea):

    def __init__(self, parent, context: Context, plugins: Plugins):
        super(__class__, self).__init__(parent)
        self._context = context
        self._tab_id = uuid.uuid4().hex

        self._logger = context.logger()
        self._plugins = plugins

        self._frames = CodecFrames(self, context, self._tab_id, plugins)

        self._main_frame = QFrame(self)
        self._main_frame_layout = QVBoxLayout()
        self._main_frame_layout.setAlignment(Qt.AlignTop)
        self._main_frame_layout.addWidget(self._frames)
        self._main_frame_layout.addWidget(VSpacer(self))
        self._main_frame.setLayout(self._main_frame_layout)
        self._frames.newFrame("", "", 0, StatusWidget.DEFAULT)

        self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
        self.setWidgetResizable(True)
        self.setWidget(self._main_frame)

    # ------------------------------------------------------------------------------------------------------------------

    def id(self):
        return self._tab_id

    def frames(self) -> CodecFrames:
        return self._frames

    # ------------------------------------------------------------------------------------------------------------------

    def toDict(self) -> List[dict]:
        return self._frames.toDict()



from dpp.core.plugin import DecoderPlugin

class Plugin(DecoderPlugin):
    """
    Decodes a URL. Plus-signs are decoded to spaces.

    Example:

        Input:
            abcdefghijklmnopqrstuvwxyz \\
            %0A%5E%C2%B0%21%22%C2%A7%24%25%26/%28%29%3D%3F%C2%B4%60%3C%3E%7C+%2C.-%3B%3A_%23%2B%27%2A%7E%0A \\
            0123456789

        Output:
            abcdefghijklmnopqrstuvwxyz
            ^°!"§$%&/()=?´`<>| ,.-;:_#+'*~
            0123456789

    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('URL+', "Thomas Engel", ["urllib"], context)

    def run(self, text):
        import urllib.parse
        return urllib.parse.unquote_plus(text)

    def can_decode_input(self, input):
        """
        Checks whether input can be decoded. When the input contains no plus sign we return False.
        When decoded input does not match the initial input some decoding must have happened so we return True.
        """
        if input and "+" in input:
            try:
                return self.run(input) != input.replace('+', ' ')
            except:
                return False
        return False



# vim: ts=8:sts=8:sw=8:noexpandtab
#
# This file is part of Decoder++
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import List

from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import QFrame, QHBoxLayout, QPushButton

from dpp.core.plugin import DecoderPlugin


class SmartDecodeButton(QFrame):
    """ A button which provides a smart-decode functionality. """

    clicked = pyqtSignal()

    def __init__(self, parent, plugins: List[DecoderPlugin]):
        super(__class__, self).__init__(parent)
        self._plugins = plugins
        layout = QHBoxLayout()
        layout.setContentsMargins(0, 6, 0, 0)
        self._button = self._init_button()
        layout.addWidget(self._button)
        self.setLayout(layout)

    def _init_button(self):
        button = QPushButton("Smart decode")
        button.clicked.connect(self._button_click_event)
        return button

    def _button_click_event(self):
        """ Forwards the "Smart decode" button click event. """
        self.clicked.emit()

    def _test_decoders(self, input, possible_decoders) -> List[DecoderPlugin]:
        """ Returns a list of decoders which did not threw an error when executing them with the specified input. """
        decoders_without_error = []
        for decoder in possible_decoders:
            try:
                decoder.select(input)
                decoders_without_error.append(decoder)
            except:
                pass
        return decoders_without_error

    def get_possible_decoders(self, input) -> List[DecoderPlugin]:
        """ Returns a list of possible decoders for the specified input. """
        possible_decoders = []
        for plugin in self._plugins:
            if plugin.can_decode_input(input):
                possible_decoders.append(plugin)
        return self._test_decoders(input, possible_decoders)



from dpp.core.plugin import DecoderPlugin

class Plugin(DecoderPlugin):
    """
    Decodes JSON Web Tokens.

    Example:

        Input:
            eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJzb21lIjoicGF5bG9hZCJ9.4twFt5NiznN84AWoo1d7KO1T_yoc0Z6XOpOVswacPZg

        Output:
            {'some': 'payload'}
    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('JWT', "Thomas Engel", ["jwt"], context)

    def run(self, text):
        import jwt
        return str(jwt.decode(text.encode('utf-8', errors='surrogateescape'), verify=False))

    def can_decode_input(self, input):
        if input and input.startswith("ey"):
            try:
                self.run(input)
                return True
            except:
                pass
        return False



from dpp.core.plugin import ScriptPlugin


class Plugin(ScriptPlugin):
    """
    Unescapes the input string.
    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('Unescape String', "Thomas Engel", [], context)

    def run(self, text):
        return text.encode('utf-8').decode('unicode_escape')



from dpp.core.plugin import HasherPlugin

class Plugin(HasherPlugin):
    """
    Hashes a string using KECCAK 256.

    Example:

        Input:
            abcdefghijklmnopqrstuvwxyz
            ^°!"§$%&/()=?´`<>| ,.-;:_#+'*~
            0123456789

        Output:
            53205b3c714c875f1d892d9ec3e7e9194f908d5b61a744d08c32f1f0a7c94c6e
    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('KECCAK 256', "Tim Menapace", ["_pysha3"], context)

    def run(self, text):
        import _pysha3
        return _pysha3.keccak_256(text.encode('utf-8', errors='surrogateescape')).hexdigest()


from dpp.core.plugin import HasherPlugin

class Plugin(HasherPlugin):
    """
    Hashes a string using KECCAK 384.

    Example:

        Input:
            abcdefghijklmnopqrstuvwxyz
            ^°!"§$%&/()=?´`<>| ,.-;:_#+'*~
            0123456789

        Output:
            2f1b4db7016471554160335867949a2d8a8bd68a002b0e0289f119ee25ab719e40ed2e8fc2f5604d8c4272ca3487cfe7
    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('KECCAK 384', "Thomas Engel", ["_pysha3"], context)

    def run(self, text):
        import _pysha3
        return _pysha3.keccak_384(text.encode('utf-8', errors='surrogateescape')).hexdigest()



from dpp.core.plugin import EncoderPlugin

class Plugin(EncoderPlugin):
    """
    Encodes a text using URL-safe BASE64.

    URL-safe BASE64 is a BASE64 derivation which remaps following characters:

     Location  Base64  Http64
     --------  ------  ------
       62        +       -
       63        /       _

    Example:

        Input:
            abcdefghijklmnopqrstuvwxyz
            ^°!"§$%&/()=?´`<>| ,.-;:_#+'*~
            0123456789

        Output:
            YWJjZGVmZ2hpamtsbW5vcHFyc3R1dnd4eXoKXsKwISLCpyQlJi8oKT0_wrRgPD58ICwuLTs6XyMrJyp-CjAxMjM0NTY3ODk=
    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('BASE64 (URL-safe)', "Thomas Engel", ["base64"], context)

    def run(self, text):
        import base64
        return base64.urlsafe_b64encode(text.encode('utf-8', errors="surrogateescape"))\
            .decode('utf-8', errors="surrogateescape")



# vim: ts=8:sts=8:sw=8:noexpandtab
#
# This file is part of Decoder++
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import collections
import copy
import importlib
import importlib.util
import json
import os
import sys
from logging import Logger
from typing import List


class PluginType(object):
    DECODER = "Decoder"
    ENCODER = "Encoder"
    HASHER = "Hasher"
    SCRIPT = "Script"
    IDENTIFY = "Identify"


class PluginConfig(object):
    """ A customizable list of configuration options for a plugin. """

    class Option(object):

        class Label(object):

            def __init__(self, key, name):
                self.key = key
                self.name = name

            def __str__(self):
                return self.key

        class Base(object):

            def __init__(self, label, value, description, is_required=True):
                """
                :param label: the label of the option (e.g. OptionLabel("search", "Search:"), ...).
                :param value: the default value of the option  (e.g. "foo", 42, False, ...).
                :param description: the description of the option.
                :param is_required: defines whether the user needs to configure this option (default = True).
                """
                self.label = label
                self.value = value
                self.description = description
                self.is_required = is_required
                self.is_initialized = False
                self.clazz = str(self.__class__.__name__)

            def _name(self):
                return self.label.name

            def _key(self):
                return self.label.key

            def __str__(self):
                self._key()

            def __deepcopy__(self, memo):
                """ Makes a deep copy of the option while reusing callable's. """
                cls = self.__class__
                result = cls.__new__(cls)
                memo[id(self)] = result
                for k, v in self.__dict__.items():
                    setattr(result, k, v if callable(v) else copy.deepcopy(v, memo))
                return result

            name = property(fget=_name)
            key = property(fget=_key)

        class String(Base):

            def __init__(self, label, value, description, is_required):
                """
                :param value: the string (e.g. "ab cd ef gh ij kl mn op qr st uv wx yz").
                """
                super(PluginConfig.Option.String, self).__init__(label, value, description, is_required)

        class Integer(Base):

            def __init__(self, label, value, description, is_required, range=None):
                """
                :param value: the integer (e.g. ..., -2, -1, 0, 1, 2, ...).
                :param range: a list containing the minimum and maximum (e.g. [-2, 2])
                """
                super(PluginConfig.Option.Integer, self).__init__(label, value, description, is_required)
                self.range = range

        class Boolean(Base):

            def __init__(self, label, value, description, is_required):
                """
                :param value: the boolean value (e.g. True/False).
                """
                super(PluginConfig.Option.Boolean, self).__init__(label, value, description, is_required)

            def _is_checked(self):
                return bool(self.value)

            is_checked = property(_is_checked)

        class Group(Boolean):
            """ A option with group name and checked status. """

            def __init__(self, label, value, description, is_required, group_name):
                """
                :param group_name: defines whether the option is associated with another group of options.
                """
                super(PluginConfig.Option.Group, self).__init__(label, value, description, is_required)
                self.group_name = group_name

        class Builder:
            """ Builds the plugin config from a dictionary. """

            def build(self, config: dict) -> dict:
                """
                Building the plugin configuration requires special care since we need to map the individual
                plugin option types (e.g. String, Boolean, ...).
                :param config: the plugin config as dictionary.
                :return: the plugin config as dictionary with options resolved to individual classes.
                """
                result = {}
                for name, value in config.items():
                    if isinstance(value, PluginConfig.Option.Base):
                        result[name] = copy.deepcopy(value)
                        continue
                    try:
                        clazz = value.pop("clazz")
                        value["label"] = PluginConfig.Option.Label(value["label"]["key"], value["label"]["name"])
                        value.pop("is_initialized")
                        result[name] = {
                            "String": PluginConfig.Option.String,
                            "Integer": PluginConfig.Option.Integer,
                            "Boolean": PluginConfig.Option.Boolean,
                            "Group": PluginConfig.Option.Group
                        }.get(clazz)(**value)
                    except:
                        raise Exception("Error while loading plugin configuration!")
                return result

    def __init__(self):
        self._config = {}
        self._validators = {}

    def add(self, option: Option.Base, validator=None):
        """
        Adds an option to the plugin configuration.

        :param option: the option to store.
        :param validator: by default validator is None which indicates that no validation is performed.
                          otherwise a callback function is expected which should be able to take three arguments,
                          the config, the codec and the value. the callback is expected to return None when validation
                          succeeded or a string containing the error message when validation failed.
        """
        self._config[option.key] = option
        self._validators[option.key] = validator

    def get(self, label: Option.Label) -> Option:
        """ Returns the option with the specified name. """
        if isinstance(label, PluginConfig.Option.Label):
            return self._config[label.key]
        else:
            return self._config[label]

    def value(self, label: Option.Label) -> str:
        """ Returns the value of the option with the specified name. """
        if isinstance(label, PluginConfig.Option.Label):
            return self._config[label.key].value
        else:
            return self._config[label].value

    def keys(self):
        """ Returns the individual configuration option names. """
        return self._config.keys()

    def items(self):
        """ Returns the individual configuration options. """
        return self._config.items()

    def count(self) -> int:
        """ Returns the number of configuration options. """
        return len(self._config.keys())

    def update(self, options, ignore_invalid=False):
        """
        Updates the value for each specified option.
        :param options: either a name-value-dictionary or an instance of PluginConfig.
        :param ignore_invalid: when set to True an exception will be thrown when illegal options are
                                discovered (default = False).

        If options is a name-value dictionary the is_initialized attribute of the associated PluginConfigOption
        is set to True, to indicate that the option was manually configured.

        Example:
            update({"foo": "bar", "bar": "foo"})
            update(config.clone())
        """
        if isinstance(options, PluginConfig):
            self._config = options._config
        else:
            def uncheck_group(group_name):
                """ Unchecks every option within the specified group. """
                for option in self._config.items():
                    if type(option) == PluginConfig.Option.Group and option.group_name == group_name:
                        option.value = False

            for option_name in options.keys():
                if option_name not in self._config:
                    if not ignore_invalid:
                        raise KeyError("Unknown plugin configuration option '{}={}'!".format(
                            option_name, options[option_name]))
                    continue

                option = self._config[option_name]

                # When option is within a group, mark it as checked and uncheck all other options
                if type(option) == PluginConfig.Option.Group:
                    uncheck_group(option.group_name)

                option.value = options[option_name]

                # Mark option as initialized
                option.is_initialized = True

    def clone(self) -> 'PluginConfig':
        """ Returns a deep copy of the plugin configuration. """
        plugin_config = PluginConfig()
        plugin_config._config = copy.deepcopy(self._config)
        plugin_config._validators = self._validators
        return plugin_config

    def validate(self, option: Option.Base, codec, input) -> str:
        """ Returns True if validation succeeded, else error message. """
        if option.key in self._validators and self._validators[option.key] is not None:
            message = self._validators[option.key](self, codec, input)
            if message is not None:
                return message
        return True

    def __str__(self):
        return json.dumps(self._config, default=lambda o: o.__dict__, sort_keys=True, indent=4)

    def toDict(self):
        """ Returns the plugin configuration as dictionary. """
        return copy.deepcopy(self._config)

    def toJSON(self):
        """ Returns the json representation of the configuration. """
        return json.dumps(self._config, default=lambda o: o.__dict__, sort_keys=True, indent=4)


class AbstractPlugin(object):
    """ Base-class to all plugins. Should not be used directly. Instead use one of its abstract implementations. """

    def __init__(self, name: str, type: str, author: str, dependencies: List[str], context: 'core.context.Context'):
        """
        Initializes a plugin.
        :param name: the name of the plugin.
        :param type: the type of the plugin (either DECODER, ENCODER, HASHER, SCRIPT, IDENTIFY).
        :param author: the author of the plugin.
        :param dependencies: the dependencies of the plugin (either None or a list of strings).
        :param context: the application context.
        """
        self._name = name
        self._safe_name, _ = os.path.splitext(os.path.basename(sys.modules[self.__class__.__module__].__file__))
        self._method_name = self._safe_name[:self._safe_name.rfind("_")]
        self._type = type
        self._author = author
        self._dependencies = dependencies
        self.config = PluginConfig()
        self._context = context
        self._aborted = False

    def setup(self, config: dict):
        """ Injects a given configuration into the plugin. """
        for name, value in config.items():
            self.config.add(value)

    def logger(self) -> Logger:
        """ Returns an logger instance which can be used within the plugin. """
        return self._context.logger()

    def is_configurable(self) -> bool:
        """
        Returns whether the plugin can be configured.
        :return: True, when configurable, otherwise False.
        """
        return self.config.count() > 0

    def is_unconfigured(self) -> List[str]:
        """
        :returns all required options which are currently not configured.
        """
        configured_groups = []
        for key in self.config.keys():
            option = self.config.get(key)
            if type(option) == PluginConfig.Option.Group and option.is_required and option.is_initialized:
                configured_groups.append(option.group_name)

        unconfigured_options = []
        for key in self.config.keys():
            option = self.config.get(key)
            if option.is_required and not option.is_initialized:
                if type(option) == PluginConfig.Option.Group and option.group_name not in configured_groups:
                    unconfigured_options.append(key)

        return unconfigured_options

    def name(self, safe_name=False) -> str:
        """
        :param safe_name: when False a human readable name is returned (e.g. "URL+"). Otherwise the name is parsed
        from the file name (e.g. url_plus) of the plugin. Defaults to False.
        :returns the name of the plugin.
        """
        if not safe_name:
            return self._name
        else:
            return self._safe_name

    def full_name(self) -> str:
        """ :returns the full name of the plugin including name and type (e.g. "URL+" -> "URL+-Decoder"). """
        return "{}-{}".format(self.name(), self.type())

    def method_name(self) -> str:
        """ :returns the safe name of the plugin without type-information (e.g. url_plus). """
        return self._method_name

    def type(self) -> str:
        """ :returns the type of the plugin (either DECODER, ENCODER, HASHER, SCRIPT, IDENTIFY). """
        return self._type

    def title(self) -> str:
        """ :returns the title of the plugin which is usually displayed to the user. """
        return "{} {}".format(self._name, self._type.capitalize())

    def author(self) -> str:
        """ :returns the author of the plugin. """
        return self._author

    def check_dependencies(self) -> List[str]:
        """
        Checks whether all specified dependencies can be loaded.
        :returns a list of unresolved dependencies, or an empty list if all dependencies could be resolved.
        """
        self._context.logger().debug("Checking dependencies for {} {}".format(self.name(), self.type()))
        unresolved_dependencies = []
        if self._dependencies:
            for dependency in self._dependencies:
                if not self.check_dependency(dependency):
                    unresolved_dependencies.append(dependency)
        return unresolved_dependencies

    def check_dependency(self, dependency) -> bool:
        """
        Checks whether the given dependency is met.
        :returns True when given dependency is met, otherwise False.
        """
        try:
            if self._context.checkDependency(dependency):
                return True
            if importlib.util.find_spec(dependency) is not None:
                return True
            importlib.import_module(dependency)
            return True
        except Exception as e:
            return False

    def dependencies(self) -> List[str]:
        """ :returns all dependencies in a list. """
        return self._dependencies

    def run(self, text: str) -> str:
        """ The main method of the plugin which must be implemented by the plugin. """
        raise NotImplementedError("Method must be implemented from upper class")

    def _run_lines(self, text: str, callback):
        """ Helper method which executes a callback for each line of text. """
        lines = []
        for text_line in text.splitlines():
            result = []
            for text_part in text_line.split(" "):
                if text_part:
                    result.append(callback(text_part))
            lines.append(' '.join(result))
        return os.linesep.join(lines)

    def select(self, text: str) -> str:
        """
        This method is usually called when the plugin gets selected for execution.
        In its simplest form it may just call the run method. But it can also be used to ask the user for additional
        parameters (e.g. by displaying dialogs).
        """
        return self.run(text)

    def _join_options_as_human_readable_string(self, options: List[str]):
        """
        Returns the list of options as human readable string.

        Examples:
            [] => ''
            ['a'] => 'a'
            ['a', 'b'] => 'a and b'
            ['a', 'b', 'c'] => 'a, b and c'

        :param options: a list of options.
        :returns the list of options as human readable string.
        """
        if not options:
            return ''
        elif len(options) == 1:
            return options[0]
        else:
            return ' and '.join([','.join(options[:-1]), options[-1]])

    def is_enabled(self) -> bool:
        """ :returns whether the plugin is enabled/disabled. """
        return self._context.config.getPluginStatus(self.full_name())

    def is_runnable(self) -> bool:
        """ :returns whether the plugin can be run. Usually true, except for NullPlugin. """
        return True

    def set_enabled(self, status):
        """ Sets the status of the plugin to enabled/disabled. """
        self._context.config.setPluginStatus(self.full_name(), status)

    def set_aborted(self, status):
        """ Sets whether the execution of the plugin was aborted by the user. """
        self._aborted = status

    def was_aborted(self) -> bool:
        """ :returns whether the execution of the plugin was aborted by the user. """
        return self._aborted

    def _set(self, name, config):
        """ Sets a fields value according to its configuration. """
        setattr(self, name, config[name])

    def __key(self):
        return (self._name, self._type)

    def __eq__(x, y):
        return x.__key() == y.__key()

    def __hash__(self):
        return hash(self.__key())

    def __deepcopy__(self, memo):
        """ Makes a deep copy of the option while reusing callable's. """
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            setattr(result, k, v if callable(v) else copy.deepcopy(v, memo))
        return result

    def toDict(self):
        return {
            "name": self.name(),
            "type": self.type(),
            "config": self.config.toDict()
        }


class DecoderPlugin(AbstractPlugin):

    def __init__(self, name: str, author: str, dependencies: List[str], context: 'core.context.Context'):
        """
        Initializes a plugin.
        :param name: the name of the plugin.
        :param author: the author of the plugin.
        :param dependencies: the dependencies of the plugin (either None or a list of strings).
        :param context: the application context.
        """
        super(__class__, self).__init__(name, PluginType.DECODER, author, dependencies, context)

    def can_decode_input(self, input) -> bool:
        """
        Returns whether it might be possible to decode the specified input with this plugin.

        Override this method to implement custom "Smart decode" behaviour. See DecoderPlugin implementations for more
        information regarding this matter.

        :returns False by default.
        """
        return False


class EncoderPlugin(AbstractPlugin):

    def __init__(self, name: str, author: str, dependencies: List[str], context: 'core.context.Context'):
        """
        Initializes a plugin.
        :param name: the name of the plugin.
        :param author: the author of the plugin.
        :param dependencies: the dependencies of the plugin (either None or a list of strings).
        :param context: the application context.
        """
        super(__class__, self).__init__(name, PluginType.ENCODER, author, dependencies, context)


class HasherPlugin(AbstractPlugin):

    def __init__(self, name: str, author: str, dependencies: List[str], context: 'core.context.Context'):
        """
        Initializes a plugin.
        :param name: the name of the plugin.
        :param author: the author of the plugin.
        :param dependencies: the dependencies of the plugin (either None or a list of strings).
        :param context: the application context.
        """
        super(__class__, self).__init__(name, PluginType.HASHER, author, dependencies, context)


class ScriptPlugin(AbstractPlugin):

    def __init__(self, name: str, author: str, dependencies: List[str], context: 'core.context.Context'):
        """
        Initializes a plugin.
        :param name: the name of the plugin.
        :param author: the author of the plugin.
        :param dependencies: the dependencies of the plugin (either None or a list of strings).
        :param context: the application context.
        """
        super(__class__, self).__init__(name, PluginType.SCRIPT, author, dependencies, context)


class IdentifyPlugin(AbstractPlugin):

    def __init__(self, name: str, author: str, dependencies: List[str], context: 'core.context.Context'):
        """
        Initializes a plugin.
        :param name: the name of the plugin.
        :param author: the author of the plugin.
        :param dependencies: the dependencies of the plugin (either None or a list of strings).
        :param context: the application context.
        """
        super(__class__, self).__init__(name, PluginType.IDENTIFY, author, dependencies, context)


class NullPlugin(AbstractPlugin):
    """ Implements a plugin which can be used as a Null-Object. """

    def __init__(self, context=None):
        super(NullPlugin, self).__init__("", "", "", [], context)

    def select(self, text: str):
        pass

    def run(self, text: str):
        pass

    def is_runnable(self) -> bool:
        return False


class PluginBuilder:
    """ Builds a plugin from a configuration item. """

    def __init__(self, context: 'core.context.Context'):
        self._context = context

    def build(self, config) -> AbstractPlugin:
        """ Returns a plugin as specified within configuration item. Returns a NullPlugin on error. """
        try:
            plugin = self._context.getPluginByName(config["name"], config["type"])
            plugin.setup(PluginConfig.Option.Builder().build(config["config"]))
            return plugin
        except Exception as e:
            self._context.logger().debug("Error building plugin:")
            self._context.logger().debug("> {}".format(config))
            self._context.logger().exception(e)
            return NullPlugin(self._context)


class PluginLoader:
    """ Loads python files of type Plugin from a specified folder. """

    def __init__(self, context):
        self._context = context
        self._logger = context.logger()
        self._unresolved_dependencies = {}
        self._errors = {}
        self._plugins_path = {}

    def load(self, paths: List['str']) -> List[AbstractPlugin]:
        """
        Loads plugins from the specified paths and returns them in an ordered list.
        :param path: the paths were plugin files (.py) are found.
        :return: ordered list of plugins.
        """
        plugins = {}
        for path in paths:
            try:
                if not os.path.exists(path):
                    self._logger.debug("Creating plugin folder '{}' ...".format(path))
                    os.makedirs(path)
                for f in os.listdir(path):
                    if f.endswith(".py"):
                        plugin = self._load_plugin(path, f)
                        if not plugin:
                            self._logger.error("Loading plugin {} at {} failed!".format(path, f))
                            continue
                        plugins[plugin.name(safe_name=True)] = plugin
            except:
                self._logger.warning("Loading plugin folder '{}' failed!".format(path))

        return [plugins[key] for key in sorted(plugins.keys())]

    def _load_plugin(self, path, f):
        self._logger.debug("Loading plugin at {}/{}".format(path, f))
        sys.path.insert(0, path)
        plugin = None
        try:
            fname, ext = os.path.splitext(f)
            mod = __import__(fname)
            plugin = mod.Plugin(self._context)
        except Exception as e:
            self._logger.error("{}: {}".format(f, str(e)))
            self._errors[f] = str(e)
            pass
        sys.path.pop(0)
        return plugin

    def _get_plugins(self):
        plugins = {}
        for path in self._plugins_path:
            for fname in self._plugins_path[path]:
                plugins[fname] = self._plugins_path[path][fname]
        return plugins


class Plugins:
    """ Defines a list of plugins and additional helper methods for working with them. """

    def __init__(self, plugin_paths: List['str'], context: 'core.context.Context'):
        self._context = context
        self._logger = context.logger()
        self._plugin_loader = PluginLoader(context)
        self._plugin_list = self._plugin_loader.load(plugin_paths)
        self._index = 0

    def names(self, type: str=None, author: str=None, safe_names=False) -> List[str]:
        """
        Returns the plugin names in a list. Does match cases.
        :param type: Filter plugin names by type (e.g. PluginType.DECODER, ...)
        :param author: Filter plugin names by author (e.g. Thomas Engel, ...)
        :param safe_names: when False human readable names are returned (e.g. "URL+"). Otherwise names are parsed
        from the file name (e.g. url_plus_encoder) of the individual plugin. Defaults to False.
        """
        if type and author:
            return [plugin.name(safe_names) for plugin in self._plugin_list if plugin.type() == type and plugin.author() == author]
        elif type:
            return [plugin.name(safe_names) for plugin in self._plugin_list if plugin.type() == type]
        elif author:
            return [plugin.name(safe_names) for plugin in self._plugin_list if plugin.author() == author]
        else:
            return [plugin.name(safe_names) for plugin in self._plugin_list]

    def types(self) -> List[str]:
        """ Returns all possible plugin types in a list. """
        return [PluginType.DECODER, PluginType.ENCODER, PluginType.HASHER, PluginType.SCRIPT, PluginType.IDENTIFY]

    def plugins(self) -> List[AbstractPlugin]:
        return self._plugin_list

    def plugin(self, name: str, type: str) -> AbstractPlugin:
        """
        Returns the plugin matching name and type. Does not match cases. There can only be one.
        :param name: The name of the plugin (e.g. SHA1/sha1).
        :param type: The type of the plugin (e.g. DECODER/decoder).
        """
        the_type = type.lower()
        the_plugin_name = name.lower()
        for plugin in self._plugin_list:
            if plugin.type().lower() == the_type and plugin.name().lower() == the_plugin_name:
                return plugin
        raise Exception("Undefined plugin '{}::{}'!".format(name, type))

    def authors(self) -> List[str]:
        """ Returns all authors in a list. """
        authors = [plugin.author() for plugin in self._plugin_list if plugin.author()]
        return [author for author, _ in collections.Counter(authors).most_common()]

    def filter(self, name: str=None, type: str=None) -> List[AbstractPlugin]:
        """
        Returns the plugins matching name and/or type. Does not match cases.
        :param name: The name of the plugin (e.g. SHA1/sha1).
        :param type: The type of the plugin (e.g. DECODER/decoder).
        :raise Exception when neither name nor type is specified.
        """
        if name and type:
            try:
                return [self.plugin(name, type)]
            except:
                return []
        if name:
            the_plugin_name = name.lower()
            return [plugin for plugin in self._plugin_list if plugin.name().lower() == the_plugin_name]
        if type:
            the_type_name = type.lower()
            return [plugin for plugin in self._plugin_list if plugin.type().lower() == the_type_name]
        raise Exception("Unknown Error '{}::{}'!".format(name, type))

    def __getitem__(self, item):
        """ Returns the specified plugin. """
        return self._plugin_list[item]

    def __len__(self):
        """ Returns the number of plugins. """
        return len(self._plugin_list)




import os

from dpp.core import Context
from dpp.core.plugin import PluginBuilder


def run_plugin(plugin_name, arguments):
    app_path = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), "dpp")
    print(app_path)
    context = Context("net.bytebutcher.decoder_plus_plus", app_path, namespace=[])
    config = {
        "name": '_'.join(plugin_name.split("_")[:-1]),
        "type": plugin_name.split("_")[-1],
        "config": {}
    }
    output = PluginBuilder(context).build(config).run(*arguments)
    print(output)
    return output




# vim: ts=8:sts=8:sw=8:noexpandtab
#
# This file is part of Decoder++
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import List

import qtawesome
from PyQt5 import QtCore
from PyQt5.QtCore import pyqtSignal, QModelIndex, QEvent
from PyQt5.QtGui import QStandardItem, QBrush
from PyQt5.QtWidgets import QVBoxLayout, QFrame, QHBoxLayout

from dpp.core import Context
from dpp.core.plugin import Plugins
from dpp.ui.widget import ListWidget, SearchField
from dpp.ui.widget.dyna_frame import DynaFrame
from dpp.ui.widget.plugin_frame import PluginFrame


class PluginSelectionFrame(QFrame):
    """ Encapsulates the SearchField and ListWidget. """

    enterPressed = pyqtSignal()
    arrowPressed = pyqtSignal()
    selectionChanged = pyqtSignal('PyQt_PyObject')
    itemChanged = pyqtSignal('PyQt_PyObject')
    textChanged = pyqtSignal('PyQt_PyObject')

    def __init__(self, context: Context, plugins: Plugins, parent=None):
        """
        Initializes the Plugin Selection Frame.
        :param context: the context of the application.
        :param plugins: the plugins to display.
        """
        super(__class__, self).__init__(parent)
        self._context = context
        self._logger = context.logger()
        self._plugins = plugins

        base_layout = QVBoxLayout()

        self._search_field = SearchField(self)
        self._search_field.setPlaceholderText("Search plugin...")
        self._search_field.setIcon(qtawesome.icon("fa.search"))
        self._search_field.enterPressed.connect(self._on_search_field_enter_key_pressed)
        self._search_field.arrowPressed.connect(self._on_search_field_arrow_pressed)
        self._search_field.textChanged.connect(self._on_search_field_text_changed)

        base_layout.addWidget(self._search_field)

        self._list_widget = ListWidget()
        self._list_widget.clicked.connect(self._show_item)
        self._list_widget.doubleClicked.connect(
            # Signals that user wishes to add this command.
            lambda: self.enterPressed.emit()
        )

        self._list_widget.dataChanged = self._on_list_widget_data_changed
        self._list_widget.selectionModel().selectionChanged.connect(self._show_item)
        self._list_widget.keyPressed.connect(self._on_list_widget_key_pressed)

        self._init_list_items()
        self._list_widget.select_item_by_index(0)

        self.itemChanged.connect(self.__on_list_widget_item_changed)

        base_layout.addWidget(self._list_widget)

        self.setLayout(base_layout)
        self.setMinimumWidth(self._list_widget.sizeHintForColumn(0) + 40)
        self.setMaximumWidth(self._list_widget.sizeHintForColumn(0) + 40)

    def _init_list_items(self):
        """ Add all plugins into a list. """
        for plugin in sorted(self._plugins, key=lambda x: getattr(x, '_name')):
            name = plugin.full_name()
            item = QStandardItem(name)
            item.setFlags(item.flags() | QtCore.Qt.ItemIsUserCheckable)
            item.setFlags(item.flags() & ~QtCore.Qt.ItemIsEditable)
            dependencies = plugin.check_dependencies()
            if dependencies:
                # Highlight plugins with unresolved dependencies
                item.setForeground(QBrush(QtCore.Qt.red, QtCore.Qt.SolidPattern))
                # Prevent user from enabling plugins with missing dependencies.
                item.setCheckable(False)

            if not dependencies and plugin.is_enabled():
                # Set the checkbox when plugins are enabled
                item.setCheckState(QtCore.Qt.Checked)
            else:
                # Unset the checkbox when plugins are enabled
                item.setCheckState(QtCore.Qt.Unchecked)
            self._list_widget.addItem(item)

    def __on_list_widget_item_changed(self, list_item: QStandardItem):
        """ Catches when items check-box was triggered. Enables/disables plugin. """
        if list_item:
            status = list_item.checkState() == QtCore.Qt.Checked
            name = list_item.text()
            plugin = self._get_item_by_name(name)
            plugin.set_enabled(status)

    def _get_item_by_name(self, name: str):
        """ Returns the plugin matching the name, otherwise None. """
        for plugin in self._plugins:
            if plugin.full_name() == name:
                return plugin
        return None

    def _get_item_names(self):
        """ Returns all plugin names in a list. """
        self._logger.debug(self._plugins[0])
        return [plugin.full_name() for plugin in self._plugins]

    def setFocus(self):
        """ Overrides the setFocus method to focus the command list widget instead of the meaningless frame. """
        self._list_widget.setFocus()

    def _show_item(self, index):
        """ Triggers the selectionChanged event with the selected command as parameter. """
        if len(self._list_widget.selectedIndexes()) > 0:
            # BUG: When list is filtered by text the emitted index does not correspond to the correct item.
            # NOTE: List uses SortFilterProxyProxyModel to implement search filter.
            # EXAMPLE:
            #   10 items in real list. 3 items in proxy list due to search filter.
            #   1st item in proxy list corresponds to 1st item in real list.
            #   2nd item in proxy list corresponds to 7th item in real list.
            #   3rd item in proxy list corresponds to 9th item in real list.
            #   When clicking the 2nd item in proxy list the text of the 7th item of the real list should be printed.
            # FIX: Map proxy index to controller index.
            item_index = self._list_widget.model().mapToSource(self._list_widget.selectedIndexes()[0]).row()
            item_name = self._list_widget.item(item_index).text()
            item = self._get_item_by_name(item_name)
            self.selectionChanged.emit(item)

    def _on_search_field_enter_key_pressed(self):
        """ When enter is pressed in the search field, the matching item is selected in the list widget. """
        self._list_widget.select_item_by_text(self._search_field.text())

    def _on_search_field_arrow_pressed(self):
        """ When an arrow-key is pressed in the search field, the list widget will gets the focus. """
        self._list_widget.setFocus(True)

    def _on_list_widget_key_pressed(self, evt: QEvent):
        """
        When a key was pressed in the list widget (except enter- and arrow-keys), the focus will be switched
        to the search field to enable the user to enter search terms without manually switching fields.
        """
        self._search_field.setFocus()
        self._search_field.setText("")
        self._search_field.keyPressEvent(evt)

    def _on_list_widget_data_changed(self, topLeft: QModelIndex, bottomRight: QModelIndex, roles: List[int]):
        """ Retriggers the item-changed event containing the corresponding item. """
        self.itemChanged.emit(self._list_widget.item(topLeft.row()))

    def _on_search_field_text_changed(self, text: str):
        """ Filters the text in the list widget based on content of search-field. """
        self._list_widget.model().setFilterRegExp(text)
        self.textChanged.emit(text)


class PluginTab(QFrame):
    """
    Frame containing a searchable check-list of plugins and a frame displaying info about the currently
    selected plugin.
    """

    enterPressed = pyqtSignal()

    def __init__(self, context, parent=None):
        """ Initializes the plugin tab. """
        super(__class__, self).__init__(parent)
        self._context = context
        self._logger = context.logger()
        self._plugins = context.plugins()
        self._plugin_selection_frame = PluginSelectionFrame(context, self._plugins, self)

        base_layout = QVBoxLayout()
        inner_widget = QFrame(self)
        inner_layout = QHBoxLayout()
        self._plugin_selection_frame.enterPressed.connect(self._on_enter_pressed)
        self._plugin_selection_frame.selectionChanged.connect(self._select_frame)

        inner_layout.addWidget(self._plugin_selection_frame)
        self._dynamic_frame = DynaFrame(self, self._get_frame_by_selection(None))
        inner_layout.addWidget(self._dynamic_frame)
        inner_widget.setLayout(inner_layout)
        base_layout.addWidget(inner_widget)

        base_layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(base_layout)
        self._plugin_selection_frame.setFocus()

    def _on_enter_pressed(self):
        """ Triggeres the enter-pressed event. """
        self.enterPressed.emit()

    def _select_frame(self, selection):
        """ Sets the widget based on the current selection. """
        self._dynamic_frame.setWidget(self._get_frame_by_selection(selection))

    def _get_frame_by_selection(self, selected_plugin=None) -> PluginFrame:
        """ Returns the frame based on the selected plugin. """
        if selected_plugin is None:
            # There should always be at least one plugin in the list.
            selected_plugin = self._plugins[0]
        return PluginFrame(self._context, selected_plugin)



from typing import List

from PyQt5.QtWidgets import QFrame, QVBoxLayout, QLabel

from dpp.core.exception import AbortedException
from dpp.ui.codec_frame import CodecFrame
from dpp.ui.widget.status_widget import StatusWidget


class CodecFrames(QFrame):

    def __init__(self, parent, context, tab_id, plugins):
        super(__class__, self).__init__(parent)
        self._context = context
        self._logger = context.logger()
        self._tab_id = tab_id
        self._plugins = plugins
        self._frames_layout = QVBoxLayout()
        self.setLayout(self._frames_layout)
        self._frames_layout.setContentsMargins(0, 0, 0, 0)
        self._init_listener(context)
        self._focused_frame = None

    # ------------------------------------------------------------------------------------------------------------------

    def _init_listener(self, context):
        self._listener = context.listener()

        def selected_frame_changed(tab_id, frame_id, input_text):
            # Always remember the currently focused frame (e.g. used for finding the currently selected frame when
            # using keyboard-shortcuts)
            if self._tab_id == tab_id:
                self._focused_frame = self.getFrameById(frame_id)

        self._listener.selectedFrameChanged.connect(selected_frame_changed)
        self._listener.textChanged.connect(
            lambda tab_id, frame_id, text: self._tab_id == tab_id and self._text_changed_event(frame_id, text))

    # ------------------------------------------------------------------------------------------------------------------

    def _text_changed_event(self, frame_id, text, is_user_action=True, do_preserve_state=False):
        frame_index = self.getFrameIndex(frame_id)
        frame = self.getFrameById(frame_id)
        if is_user_action:
            frame._status_widget.setStatus("DEFAULT")
        while frame:
            if do_preserve_state and self.hasNextFrame(frame_index) and \
                    self.getFrameByIndex(frame_index + 1)._status_widget.hasStatus("DEFAULT"):
                # Do not overwrite content of frames which are in default-state.
                # Usually done when moving frames to a new position whereby custom user-input should not be
                # overwritten.
                frame = self.getFrameByIndex(frame_index + 1)
                continue

            input_text = frame.getInputText()
            frame.header().refresh()
            plugin = frame._combo_box_frame.selectedPlugin()
            if not plugin.is_runnable():
                break

            frame = self._execute_plugin_run(frame.id(), input_text, plugin)

    # ------------------------------------------------------------------------------------------------------------------

    def _execute_plugin_run(self, frame_id, input_text, plugin) -> CodecFrame:
        frame_index = self.getFrameIndex(frame_id) + 1
        output_text = ""
        status = StatusWidget.SUCCESS
        error = None
        try:
            output_text = plugin.run(input_text)
        except BaseException as e:
            status = StatusWidget.ERROR
            error = str(e)
            self._logger.error('{} {}: {}'.format(plugin.name(), plugin.type(), error))

        return self.newFrame(output_text, plugin.title(), frame_index, status=status, msg=error)

    def _execute_plugin_select(self, frame_id, input_text, plugin):
        frame_index = self.getFrameIndex(frame_id) + 1
        try:
            plugin.set_aborted(False)
            output = plugin.select(input_text)
            status = StatusWidget.SUCCESS
            error = None
        except AbortedException as e:
            # User aborted selection. This usually happens when a user clicks the cancel-button within a codec-dialog.
            plugin.set_aborted(True)
            self._logger.debug(str(e))
            return
        except BaseException as e:
            output = ""
            status = StatusWidget.ERROR
            error = str(e)
            self._logger.error('{} {}: {}'.format(plugin.name(), plugin.type(), error))
            self._logger.exception(error, exc_info=self._context.isDebugModeEnabled())

        self.newFrame(output, plugin.title(), frame_index, status=status, msg=error).focusInputText()

    def _refill_frame(self, text, title, frame_index, status, msg=None):
        if status == StatusWidget.ERROR:
            # ERROR usually indicates that codec execution failed and there is no text to display.
            # However, we indicate to the user that there is an error (flashStatus) for this frame and any
            # following frame.
            frame = self.getFrameByIndex(frame_index)
            while frame_index < self.getFramesCount():
                _frame = self.getFrameByIndex(frame_index)
                _frame.flashStatus(status, msg)
                _frame.header().refresh()
                # Display error only for the first frame.
                msg = None
                frame_index = frame_index + 1
        else:
            # Display status and insert text into frame when there was no error.
            # Note that setInputText may result into another call to newFrame with frame_index + 1.
            frame = self.getFrameByIndex(frame_index)
            frame.setInputText(text)
            frame.header().refresh()
            frame.flashStatus(status, msg)

        return frame

    def _get_plugin_config(self, frame_id: str):
        # Remember: We press the config-button of a frame, but want the plugin-config of the previous frame ...
        frame_index = self.getFrameIndex(frame_id)
        previous_frame = self.getFrameByIndex(frame_index - 1)
        if previous_frame:
            plugin = previous_frame.getPlugin()
            input_text = previous_frame.getInputText()
            self._execute_plugin_select(previous_frame.id(), input_text, plugin)

    def _reset_frames(self, frame_id):
        _frame_index = self.getFrameIndex(frame_id)
        self._logger.debug("Reset frames after index {} up until {}".format(_frame_index, self.getFramesCount() - 1))
        for frame_index in range(self.getFramesCount() - 1, _frame_index, -1):
            self._logger.debug("Reset frame with index {}".format(frame_index))
            self.getFrameByIndex(frame_index).deleteLater()

    # ------------------------------------------------------------------------------------------------------------------

    def getFrameIndex(self, frame_id) -> int:
        frame = None
        index = 0
        for current_frame in self.getFrames():
            if current_frame.id() == frame_id:
                frame = current_frame
                break
            index = index + 1
        return index if frame else -1

    def getFrameById(self, frame_id) -> CodecFrame:
        for frame in self.getFrames():
            if frame.id() == frame_id:
                return frame

    def getFrameByIndex(self, index) -> CodecFrame:
        if self.layout().itemAt(index):
            return self.layout().itemAt(index).widget()

    def getFramesCount(self) -> int:
        return self.layout().count()

    def getFocusedFrame(self) -> CodecFrame:
        widget = self.focusWidget()
        while widget:
            if isinstance(widget, CodecFrame):
                return widget
            widget = widget.parent()
        return self.layout().itemAt(0).widget()

    def getFrames(self) -> List[CodecFrame]:
        frames = []
        for frameIndex in range(0, self.layout().count()):
            frames.append(self.layout().itemAt(frameIndex).widget())
        return frames

    def hasNextFrame(self, frame_index: int) -> bool:
        return frame_index < (self.getFramesCount() - 1)

    def hasPreviousFrame(self, frame_index: int) -> bool:
        return frame_index > 0

    # ------------------------------------------------------------------------------------------------------------------

    def newFrame(self, text, title, frame_index, status, msg=None) -> CodecFrame:
        if frame_index < self.getFramesCount():
            self._logger.debug("Refill frame at index {}".format(frame_index))
            return self._refill_frame(text, title, frame_index, status, msg)

        self._logger.debug("Add new frame")
        previous_frame = self.getFrameByIndex(frame_index - 1)
        new_frame = CodecFrame(self, self._context, self._tab_id, self, self._plugins, text)
        self.layout().addWidget(new_frame)

        new_frame.pluginSelected.connect(self._execute_plugin_select)
        new_frame.pluginDeselected.connect(self._reset_frames)
        new_frame.configButtonClicked.connect(lambda frame_id: self._get_plugin_config(frame_id))
        new_frame.upButtonClicked.connect(self.moveFrameUp)
        new_frame.downButtonClicked.connect(self.moveFrameDown)
        new_frame.closeButtonClicked.connect(self.closeFrame)

        if frame_index > 0:
            # Every new frame (except the first frame) should signal success/error.
            new_frame.flashStatus(status, msg)

        if previous_frame and frame_index > 1:
            # Auto-collapse frame where plugin was selected (except first frame and when this frame had focus).
            previous_frame.setCollapsed(not (self._focused_frame and self._focused_frame == previous_frame))

        new_frame.setContentsMargins(0, 0, 0, 0)
        new_frame.layout().setContentsMargins(0, 0, 0, 0)
        new_frame.header().refresh()
        if previous_frame: previous_frame.header().refresh()
        return new_frame

    def closeFrame(self, frame_id):
        """
        Closes/Removes the frame with the specified frame-id.
        """
        _frame_index = self.getFrameIndex(frame_id)
        self._logger.debug("Close frame {}:{}".format(_frame_index, frame_id))
        if _frame_index <= 0:
            self._logger.error("Invalid close! Can not close first or invalid frame.".format(_frame_index, frame_id))
            return

        frame = self.getFrameByIndex(_frame_index)
        previous_frame = self.getFrameByIndex(_frame_index - 1) # There is always a previous frame.
        if not self.hasNextFrame(_frame_index):
            # If this is the last frame the combo-boxes of the previous frame needs to reset.
            self.layout().removeWidget(frame) # remove frame from layout to avoid side-effects with header refresh
            frame.deleteLater()
            previous_frame.getComboBoxes().resetAll()
            # Usability: Always show the content of the last frame.
            previous_frame.setCollapsed(False)
            # Update headings since some buttons may need to be en-/disabled.
            previous_frame.header().refresh()
        else:
            # Otherwise the selected plugin of the previous frame needs to be executed.
            next_frame = self.getFrameByIndex(_frame_index + 1)
            self.layout().removeWidget(frame) # remove frame from layout to avoid side-effects with header refresh
            frame.deleteLater()
            self._execute_plugin_run(previous_frame.id(), previous_frame.getInputText(), previous_frame.getPlugin())
            # Update headings since some buttons may need to be en-/disabled.
            next_frame.header().refresh()
            previous_frame.header().refresh()

    def moveFrameUp(self, frame_id: str):
        # Example:
        #
        #   1. "Hello, world" - Base64 - open
        #   2. "<base64> of hello world" - Sha256 - collapsed
        #   3. "<sha256> of base64 hello world" - <> - open
        #
        #   Case 1. Press up on Frame 1
        #
        #       First frame can not be moved.
        #
        #   Case 2. Press down on Frame 2
        #
        #       Second frame can not be moved.
        #
        #   Case 3. Press down on Frame 3
        #
        #       Switching the codec of the previous frame with Frame 3 results in:
        #
        #       1. "Hello, world" - Sha256 - open
        #       2. "<sha256> of hello world" - Base64 - collapsed
        #       3. "<base64> of <sha256> of hello world" - open
        #
        #       Note that the text of the first frame is not touched. More precisely any frame which is in default
        #       state should not be touched. However, frames which contain an error can be overwritten.
        #
        frame_index = self.getFrameIndex(frame_id)
        codec_frame = self.getFrameByIndex(frame_index)
        if not codec_frame or frame_index < 2:
            # First and second frame can not be moved up.
            self._logger.debug('moveFrameUp({}:{}): invalid move'.format(frame_index, frame_id))
            return

        self.switchFrames(frame_index - 2, frame_index - 1)
        codec_frame = self.getFrameByIndex(frame_index - 2)
        self._text_changed_event(codec_frame.id(), codec_frame.getInputText(), is_user_action=False, do_preserve_state=True)

    def moveFrameDown(self, frame_id: str):
        # Example:
        #
        #   1. "Hello, world" - Base64 - open
        #   2. "<base64> of hello world" - Sha256 - collapsed
        #   3. "<sha256> of base64 hello world" - <> - open
        #
        #   Case 1. Press down on Frame 1
        #
        #       First frame can not be moved.
        #
        #   Case 2. Press down on Frame 2
        #
        #       Switching the codec of the previous frame with Frame 2 results in:
        #
        #       1. "Hello, world" - Sha256 - open
        #       2. "<sha256> of hello world" - Base64 - collapsed
        #       3. "<base64> of <sha256> of hello world" - open
        #
        #       Note that the text of the first frame is not touched. More precisely any frame which is in default
        #       state should not be touched. However, frames which contain an error can be overwritten.
        #
        #   Case 3. Press down on Frame 3
        #
        #       Last frame can not be moved.
        #
        frame_index = self.getFrameIndex(frame_id)
        codec_frame = self.getFrameByIndex(frame_index)
        if not codec_frame or frame_index == 0 or frame_index == self.getFramesCount() - 1:
            # First and last frame can not be moved down.
            self._logger.debug('moveFrameDown({}:{}): invalid move'.format(frame_index, frame_id))
            return

        self.switchFrames(frame_index - 1, frame_index)
        codec_frame = self.getFrameByIndex(frame_index - 1)
        self._text_changed_event(codec_frame.id(), codec_frame.getInputText(), is_user_action=False, do_preserve_state=True)

    def switchFrames(self, index1, index2):
        frame1 = self.getFrameByIndex(index1)
        plugin1 = frame1.getPlugin()
        frame2 = self.getFrameByIndex(index2)
        plugin2 = frame2.getPlugin()
        frame2.setPlugin(plugin1)
        frame1.setPlugin(plugin2)

    # ------------------------------------------------------------------------------------------------------------------

    def toDict(self) -> List[dict]:
        return [frame.toDict() for frame in self.getFrames()]


import re

from dpp.core.plugin import DecoderPlugin

class Plugin(DecoderPlugin):
    """
    Decodes a hex shell code.

    Example:

        Input:
            \\x61\\x62\\x63\\x64\\x65\\x66\\x67\\x68\\x69\\x6a\\x6b\\x6c\\x6d\\x6e \\
            \\x6f\\x70\\x71\\x72\\x73\\x74\\x75\\x76\\x77\\x78\\x79\\x7a\\x0a\\x5e \\
            \\xc2\\xb0\\x21\\x22\\xc2\\xa7\\x24\\x25\\x26\\x2f\\x28\\x29\\x3d\\x3f \\
            \\xc2\\xb4\\x60\\x3c\\x3e\\x7c\\x20\\x2c\\x2e\\x2d\\x3b\\x3a\\x5f\\x23 \\
            \\x2b\\x27\\x2a\\x7e\\x0a\\x30\\x31\\x32\\x33\\x34\\x35\\x36\\x37\\x38 \\
            \\x39

        Output:
            abcdefghijklmnopqrstuvwxyz
            ^°!"§$%&/()=?´`<>| ,.-;:_#+'*~
            0123456789
    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('HEX (shell)', "Thomas Engel", ["codecs"], context)

    def run(self, text):
        if text:
            import codecs
            import re
            output = text
            for hex_code in set(sorted(re.findall(r'\\[Xx][0-9a-fA-F][0-9a-fA-F]', text))):
                try:
                    output = output.replace(hex_code, codecs.decode(hex_code[-2:], 'hex').decode('utf-8', errors='surrogateescape'))
                except:
                    pass
            return output
        else:
            return ""

    def can_decode_input(self, input):
        contains_hex = re.findall(r'\\[Xx][0-9a-fA-F][0-9a-fA-F]', input)
        if contains_hex:
            return True
        return False



from dpp.core.plugin import EncoderPlugin

class Plugin(EncoderPlugin):
    """
    Encodes a text using Base64.

    Example:

        Input:
            abcdefghijklmnopqrstuvwxyz
            ^°!"§$%&/()=?´`<>| ,.-;:_#+'*~
            0123456789

        Output:
            YWJjZGVmZ2hpamtsbW5vcHFyc3R1dnd4eXoKXsKwISLCpyQlJi8oKT0/wrRgPD58ICwuLTs6XyMrJyp+CjAxMjM0NTY3ODk=
    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('BASE64', "Thomas Engel", ["base64"], context)

    def run(self, text):
        import base64
        return base64.b64encode(text.encode('utf-8', errors="surrogateescape")).decode('utf-8', errors="surrogateescape")



# vim: ts=8:sts=8:sw=8:noexpandtab
#
# This file is part of Decoder++
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# The code is based on https://github.com/obdasystems/eddy/blob/master/eddy/ui/dock.py
from PyQt5 import QtCore
from PyQt5 import QtGui
from PyQt5 import QtWidgets
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QTabBar, QDockWidget, QStyle, QApplication, QToolButton, QHBoxLayout, QFrame, QWidget


class DockWidget(QtWidgets.QDockWidget):
    """
    This class implements the container for docking area widgets.
    """
    def __init__(self, title: str, icon: QIcon, parent):
        """
        Initialize the dock widget.
        """
        super().__init__(title, parent, QtCore.Qt.Widget)
        self.icon = icon

        # Add customized title bar widget
        self.title_bar_widget = TitleBarWidget(title, icon, self)
        self.setTitleBarWidget(self.title_bar_widget)

        # Configure main layout
        self._dock_frame = QFrame()
        self._dock_layout = QHBoxLayout()
        self._dock_frame.setLayout(self._dock_layout)
        self.setWidget(self._dock_frame)

        # Setup event listeners
        self.dockLocationChanged.connect(lambda area: self.updateTabBarIcon())
        self.featuresChanged.connect(lambda features: self.title_bar_widget.featuresChangedEvent(features))

    #############################################
    #   Interface
    #############################################

    def addWidget(self, widget: QWidget):
        self._dock_layout.addWidget(widget)

    def addTitleBarButton(self, button):
        """
        Add a button to the right side of the titlebar of this widget.
        :type button: T <= QPushButton|QToolButton
        """
        widget = self.titleBarWidget()
        widget.addButton(button)
        widget.updateLayout()

    def updateTabBarIcon(self):
        """
        Updates the tab bar icon.
        """
        tab_bar = self.parent().findChild(QTabBar, "")
        if tab_bar:
            for index in range(0, tab_bar.count()):
                if self.windowTitle() == tab_bar.tabText(index):
                    tab_bar.setTabIcon(index, self.icon)
                    break

    def hide(self):
        """ Hides the dock. """
        self._dock_frame.setVisible(False)

    def show(self):
        """ Shows the dock. """
        self._dock_frame.setVisible(True)

    #############################################
    #   Events
    #############################################

    def closeEvent(self, QCloseEvent):
        """ Closes the dock. """
        self.hide()


class TitleBarWidget(QtWidgets.QWidget):
    """
    This class implements the title area of docking area widgets.
    """
    def __init__(self, title: str, icon: QIcon, parent):
        """
        Initialize the widget.
        :type title: str
        :type icon: QIcon
        :type parent: QDockWidget
        """
        super().__init__(parent)

        # Create title bar icon and title
        self.imageLabel = QtWidgets.QLabel(self)
        self.imageLabel.setPixmap(icon.pixmap(18))
        self.imageLabel.setAlignment(QtCore.Qt.AlignLeft|QtCore.Qt.AlignVCenter)
        self.imageLabel.setContentsMargins(0, 0, 0, 0)
        self.imageLabel.setFixedSize(18, 18)
        self.titleLabel = QtWidgets.QLabel(title, self)
        self.titleLabel.setAlignment(QtCore.Qt.AlignLeft|QtCore.Qt.AlignVCenter)
        self.titleLabel.setContentsMargins(4, 0, 0, 0)

        # Create standard buttons
        iconSize = QApplication.style().standardIcon(
            QStyle.SP_TitleBarNormalButton).actualSize(
            QtCore.QSize(100, 100))
        buttonSize = iconSize + QtCore.QSize(4, 4)

        self.dockButton = QToolButton(self)
        self.dockButton.setIcon(QApplication.style().standardIcon(
            QStyle.SP_TitleBarNormalButton))
        self.dockButton.setMaximumSize(buttonSize)
        self.dockButton.setAutoRaise(True)
        self.dockButton.clicked.connect(self.toggleFloating)

        self.closeButton = QToolButton(self)
        self.closeButton.setMaximumSize(buttonSize)
        self.closeButton.setAutoRaise(True)
        self.closeButton.setIcon(QApplication.style().standardIcon(
            QStyle.SP_DockWidgetCloseButton))
        self.closeButton.clicked.connect(self.parent().close)
        self.buttons = [self.dockButton, self.closeButton]

        # Configure layout
        self.mainLayout = QtWidgets.QHBoxLayout(self)
        self.mainLayout.setContentsMargins(0, 0, 0, 0)
        self.mainLayout.setSpacing(0)
        self.setContentsMargins(6, 4, 6, 4)
        self.setContextMenuPolicy(QtCore.Qt.PreventContextMenu)
        self.updateLayout()

    #############################################
    #   Interface
    #############################################

    def addButton(self, button):
        """
        Add a button to the right side of the titlebar, before the close button.
        :type button: T <= QPushButton|QToolButton
        """
        self.buttons.insert(0, button)

    def toggleFloating(self):
        """
        Toggles floating.
        """
        self.parent().setFloating(not self.parent().isFloating())

    def updateLayout(self):
        """
        Redraw the widget by updating its layout.
        """
        # CLEAR CURRENTY LAYOUT
        for i in reversed(range(self.mainLayout.count())):
            item = self.mainLayout.itemAt(i)
            self.mainLayout.removeItem(item)
        # DISPOSE NEW ELEMENTS
        self.mainLayout.addWidget(self.imageLabel, 0, QtCore.Qt.AlignLeft|QtCore.Qt.AlignVCenter)
        self.mainLayout.addWidget(self.titleLabel, 1, QtCore.Qt.AlignLeft|QtCore.Qt.AlignVCenter)
        for button in self.buttons:
            self.mainLayout.addWidget(button, 0, QtCore.Qt.AlignRight|QtCore.Qt.AlignVCenter)

    #############################################
    #   Events
    #############################################

    def featuresChangedEvent(self, features):
        """
        Updates the features of the title bar widget.
        :param features: const
        """
        self.dockButton.setVisible(features & QDockWidget.DockWidgetFloatable)
        self.closeButton.setVisible(features & QDockWidget.DockWidgetClosable)

    def mouseDoubleClickEvent(self, mouseEvent):
        """
        Executed when the mouse is double clicked on the widget.
        :type mouseEvent: QMouseEvent
        """
        pass

    def paintEvent(self, paintEvent):
        """
        This is needed for the widget to pick the stylesheet.
        :type paintEvent: QPaintEvent
        """
        option = QtWidgets.QStyleOption()
        option.initFrom(self)
        painter = QtGui.QPainter(self)
        style = self.style()
        style.drawPrimitive(QtWidgets.QStyle.PE_Widget, option, painter, self)


"""Default execution entry point if running the package via python -m."""
from dpp import runner
import sys


def main():
    """Run dpp from script entry point."""
    return runner.main()


if __name__ == '__main__':
    sys.exit(main())


# vim: ts=8:sts=8:sw=8:noexpandtab
#
# This file is part of Decoder++
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import ctypes
import datetime
import os
from typing import List
import qtawesome

from PyQt5.QtCore import QEvent, Qt
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QMainWindow, QDockWidget, QTabBar

from dpp.core.logging import LogEntry, LogFilter
from dpp.ui import IconLabel
from dpp.ui.dialog.config_dialog import ConfigDialog
from dpp.ui.dock.hex_dock import HexDock
from dpp.ui.dock.log_dock import LogDock
from dpp.ui.widget.message_widget import MessageWidget


class MainWindow(QMainWindow):

    def __init__(self, context: 'core.context.Context', input: str = None):
        super().__init__()
        self._context = context
        self._logger = context.logger()

        #############################################
        #   Initialize status bar
        #############################################

        self._message_widget = self._init_message_widget()
        self._logger.handlers[0].addFilter(self._init_log_filter(self._message_widget))
        self.statusBar().addWidget(self._message_widget)
        self.statusBar().addPermanentWidget(self._init_hidden_dialog())

        #############################################
        #   Initialize docks
        #############################################

        # Initialize empty dock
        self._empty_dock = QDockWidget("", self)
        self._empty_dock.setFeatures(QDockWidget.NoDockWidgetFeatures)
        self.addDockWidget(Qt.BottomDockWidgetArea, self._empty_dock)
        self._empty_dock.visibilityChanged.connect(self.show_dock)

        # Initialize log dock
        self._log_entries = []
        self._log_dock = LogDock(self._log_entries, self)
        self._log_dock.clearEvent.connect(self._message_widget.resetCount)
        self._log_dock.setFeatures(QDockWidget.DockWidgetFloatable)
        self.addDockWidget(Qt.BottomDockWidgetArea, self._log_dock)
        self.tabifyDockWidget(self._empty_dock, self._log_dock)
        self._log_dock.visibilityChanged.connect(self.show_dock)

        # Initialize hex dock
        self._hex_dock = HexDock(self._context, self)
        self._hex_dock.setFeatures(QDockWidget.DockWidgetFloatable)
        self.addDockWidget(Qt.BottomDockWidgetArea, self._hex_dock)
        self.tabifyDockWidget(self._empty_dock, self._hex_dock)
        self._hex_dock.visibilityChanged.connect(self.show_dock)

        # Initialize dock tab bar
        self._dock_tab_bar = self.findChild(QTabBar, "")
        self._dock_tab_bar.tabBarDoubleClicked.connect(lambda e: self._empty_dock.raise_())

        # Do only dock tab bar on startup
        self._empty_dock.raise_()

        #############################################
        #   Initialize window
        #############################################

        self._init_window_size()
        self.setWindowTitle("Decoder++")
        self.setWindowIcon(QIcon(os.path.join(self._context.getAppPath(), 'images', 'dpp_128.png')))

    def _init_log_filter(self, message_widget):
        """ Initializes the log filter which catches log events to be shown in the statusbar. """
        log_filter = LogFilter(self)
        log_filter.logInfoEvent.connect(message_widget.showInfo)
        log_filter.logErrorEvent.connect(message_widget.showError)
        log_filter.logDebugEvent.connect(message_widget.showDebug)
        return log_filter

    def _init_window_size(self):
        """ Initializes the window size. Looks and uses any previously saved sizing. """
        size = self._context.config.getSize()
        if size:
            self.resize(self._context.config.getSize())
        self.setMinimumWidth(520)
        self.setMinimumHeight(300)

    def _init_message_widget(self) -> MessageWidget:
        """ Inits the message widget located in the statusbar. """
        message_widget = MessageWidget()
        message_widget.messageReceived.connect(self._log_message)
        message_widget.infoClicked.connect(lambda: self._toggle_log_dock(LogDock.Filter.INFO))
        message_widget.errorClicked.connect(lambda: self._toggle_log_dock(LogDock.Filter.ERROR))
        message_widget.messageClicked.connect(lambda: self._toggle_log_dock(LogDock.Filter.INFO, LogDock.Filter.ERROR))
        return message_widget

    def _init_hidden_dialog(self):
        """ Inits the icon which opens the hidden dialog on mouse press. """
        about_label = IconLabel(self, QIcon(os.path.join(self._context.getAppPath(), 'images', 'hidden.png')))
        about_label.mousePressEvent = lambda e: self._show_hidden_dialog()
        return about_label

    def _log_message(self, type: str, message: str):
        """
        Adds the message to the log dock.
        :param type: the type of the message (e.g. info, error, debug).
        :param message: the message.
        """
        now = datetime.datetime.now().time()
        log_entry = LogEntry(
            "{hour:02d}:{minute:02d}:{second:02d}".format(hour=now.hour, minute=now.minute, second=now.second), type,
            message)
        if self._log_dock:
            self._log_dock.addItem(log_entry)
        self._log_entries.append(log_entry)

    def _show_hidden_dialog(self):
        """ Shows the hidden dialog. """
        hidden_dialog = ConfigDialog(self, self._context)
        hidden_dialog.exec_()

    def _toggle_log_dock(self, *filters: List[str], **kwargs):
        """ Shows/Hides the log dock. """
        is_log_dock_visible = not self._log_dock.visibleRegion().isEmpty()
        if not is_log_dock_visible:
            self._log_dock.raise_()
            for filter in self._log_dock.getFilters():
                self._log_dock.setFilterChecked(filter, filter in filters)
        else:
            self._empty_dock.raise_()

    def _toggle_hex_dock(self):
        """ Shows/Hides the hex dock. """
        is_hex_dock_visible = not self._hex_dock.visibleRegion().isEmpty()
        if not is_hex_dock_visible:
            self._hex_dock.raise_()
        else:
            self._empty_dock.raise_()


    def show_dock(self, e: QEvent):
        """ Shows/Hides the docks. """
        is_empty_dock_visible = not self._empty_dock.visibleRegion().isEmpty()
        for dock in [self._log_dock, self._hex_dock]:
            if is_empty_dock_visible:
                if not dock.isFloating():
                    dock.hide()
            else:
                dock.show()

    def closeEvent(self, e: QEvent):
        """ Closes the main window and saves window-size and -position. """
        self._context.config.setSize(self.size())
        self._context.config.setPosition(self.pos())
        e.accept()

    def setWindowIcon(self, icon: QIcon):
        """ Sets the window icon of the main window. """
        super(__class__, self).setWindowIcon(icon)
        if os.name == 'nt':
            # Set explicit app id to show decoder-plus-plus icon in the taskbar.
            ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(self._context.getAppID())



from dpp.core.plugin import HasherPlugin

class Plugin(HasherPlugin):
    """
    Hashes a string using SHA3 256.

    Example:

        Input:
            abcdefghijklmnopqrstuvwxyz
            ^°!"§$%&/()=?´`<>| ,.-;:_#+'*~
            0123456789

        Output:
            15fd91bcfe6b705f56b436766dcd01e2d3bda8156154c80a6e030f2515956ff8
    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('SHA3 256', "Thomas Engel", ["_pysha3"], context)

    def run(self, text):
        import _pysha3
        return _pysha3.sha3_256(text.encode('utf-8', errors='surrogateescape')).hexdigest()



from dpp.core.plugin import EncoderPlugin

class Plugin(EncoderPlugin):
    """
    Encodes a string to HTML.

    Example:

        Input:
            abcdefghijklmnopqrstuvwxyz
            ^°!"§$%&/()=?´`<>| ,.-;:_#+'*~
            0123456789

        Output:
            abcdefghijklmnopqrstuvwxyz
            ^°!&quot;§$%&amp;/()=?´`&lt;&gt;| ,.-;:_#+&#x27;*~
            0123456789
    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('HTML', "Thomas Engel", ["html"], context)

    def run(self, text):
        import html
        return html.escape(text)



# vim: ts=8:sts=8:sw=8:noexpandtab
#
# This file is part of Decoder++
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse


class SingleArgs(argparse.Action):
    """
    Defines an Argparse-Action which allows specifying an argument only once.

    Example:

        parser = argparse.ArgumentParser()
        parser.add_argument('--test1', action=SingleArgs)
        parser.add_argument('--test2', action=SingleArgs)

        parser.parse_args(['--test2', '2', '--test1', '1', '--test2', '3'])
        argparse.ArgumentTypeError: Argument "test2" can only be used once!

    """

    def __call__(self, parser, args, values, option_string=None):
        if getattr(args, self.dest) is not None:
            raise argparse.ArgumentTypeError('Argument "{f}" can only be used once!'.format(f=self.dest))
        setattr(args, self.dest, values)



from dpp.core.plugin import ScriptPlugin


class Plugin(ScriptPlugin):
    """
    This little beautifier will reformat and re-indent bookmarklets, ugly JavaScript, unpack scripts packed by
    Dean Edward’s popular packer, as well as partly deobfuscate scripts processed by the npm package
    javascript-obfuscator.
    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('JS Beautify', "Thomas Engel", ["jsbeautifier"], context)

    def run(self, text):
        import jsbeautifier
        return jsbeautifier.beautify(text)



#!/usr/bin/env python3
# vim: ts=8:sts=8:sw=8:noexpandtab
#
# This file is part of Decoder++
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os
import signal
import sys
import argparse
from collections import namedtuple
from typing import List

# Load fuzzywuzzy while ignoring unnecessary warning about missing levenstein package.
import warnings

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from fuzzywuzzy import process

from PyQt5.QtWidgets import QApplication


# FIX #27: Add 'dpp' to package path if not present. 
#          This may happen when dpp was not installed via setup.py.
DPP_PACKAGE_PATH = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
if DPP_PACKAGE_PATH not in sys.path:
    sys.path.append(DPP_PACKAGE_PATH)


from dpp.core.argparse.ordered_multi_args import OrderedMultiArgs
from dpp.core.argparse.single_args import SingleArgs
from dpp.core.context import Context
from dpp.core.decoder_plus_plus import Decoder, Encoder, Hasher, Script, DecoderPlusPlus, Identify
from dpp.core.plugin import PluginType

from dpp.ui.decoder_plus_plus_gui import DecoderPlusPlusDialog, DecoderPlusPlusWindow
from dpp.ui.instance_handler import InstanceHandler

# Abort program execution on ctrl+c
signal.signal(signal.SIGINT, signal.SIG_DFL)


def init_builder(context: 'dpp.core.context.Context'):

    def _init_builder(plugin: 'dpp.core.plugin.plugins.PluginHolder', clazz):
        def list(self, filter_terms=()) -> List[str]:
            codecs = [method for method in dir(self) if not method.startswith("_") and
                    method not in ["list", "decode", "encode", "hash", "script", "run"]]
            return [codec for codec in codecs if all(filter_term in codec for filter_term in filter_terms)]

        # Add list method to clazz.
        setattr(clazz, "list", list)

        # Add plugins to clazz.
        def runner(plugin):

            def sys_exit(exit_code):
                """ Wraps the sys.exit call. Only sys.exit in command line mode. """
                if context.mode() == Context.Mode.COMMAND_LINE:
                    sys.exit(exit_code)

            def show_help():
                """ Shows the plugin config options. """

                def max_length(attr, title):
                    """ :returns the maximum string length of a specific plugin option attribute. """
                    lens = [len(title)]
                    for key in plugin.config.keys():
                        if hasattr(plugin.config.get(key), attr):
                            lens.append(len(str(getattr(plugin.config.get(key), attr))))
                    return max(lens)

                # Print title
                print()
                print(plugin.name())
                print(len(plugin.name()) * '=')
                print()

                # Print header
                name_max_length = max_length("key", "Name")
                value_max_length = max_length("value", "Value")
                group_max_length = max_length("group_name", "Group")
                row_format = "{:>" + str(name_max_length) + "}  " + \
                             "{:<" + str(value_max_length) + "}  " + \
                             "{:<" + str(group_max_length) + "}  " + \
                             "{:<8}  " + \
                             "{}"
                print(row_format.format("Name", "Value", "Group", "Required", "Description"))
                print(row_format.format("----", "-----", "-----", "--------", "-----------"))

                # Print rows
                for key in plugin.config.keys():
                    option = plugin.config.get(key)
                    if type(option.value) == type(True):
                        value = "True" if bool(option.value) else "False"
                    else:
                        value = option.value
                    group_name = option.group_name if hasattr(option, "group_name") else ""
                    is_required = "yes" if option.is_required else "no"
                    print(row_format.format(option.key, value, group_name, is_required, option.description))

                print()

            def natural_join(list):
                """ Joins a list of strings (e.g. ["1", "2", "3"] => "1, 2 and 3"). """
                if not list:
                    return ""
                elif len(list) == 1:
                    return list[0]
                else:
                    return " and ".join([", ".join(list[:-1]), list[-1]])

            def _runner(self, **kwargs):

                plugin.config.update(kwargs, ignore_invalid=True)

                if "help" in kwargs.keys():
                    show_help()
                    return sys_exit(0)

                if plugin.is_configurable() and plugin.is_unconfigured():
                    invalid_plugin_options = [key for key in kwargs if key not in plugin.config.keys()]
                    if invalid_plugin_options:
                        invalid_plugin_option = invalid_plugin_options[0]
                        suggestions = [result for result in
                                       process.extract(invalid_plugin_option, plugin.config.keys())]
                        suggestion = 'Did you mean "{}"?'.format(suggestions[0][0]) if suggestions else ""
                        context.logger().error("Can not run '{}'! Invalid option {}. {}".format(
                            plugin.name(safe_name=True), invalid_plugin_option, suggestion))
                        return sys_exit(1)

                unconfigured_plugin_options = plugin.is_unconfigured()
                if unconfigured_plugin_options:
                    context.logger().error("Can not run '{}'! Missing required option {}.".format(
                        plugin.name(safe_name=True), natural_join(unconfigured_plugin_options)))
                    return sys_exit(1)

                self._input = plugin.run(self._input)
                return self
            return _runner

        setattr(clazz, plugin.method_name(), runner(plugin))

    plugins = context.plugins()
    clazz_map = {
        PluginType.ENCODER: Encoder,
        PluginType.DECODER: Decoder,
        PluginType.HASHER: Hasher,
        PluginType.SCRIPT: Script,
        PluginType.IDENTIFY: Identify
    }
    for plugin in plugins.plugins():
        if not plugin.type() in clazz_map:
            context.logger().debug("Can not load plugin '{}'! Invalid type '{}'!".format(plugin.name(safe_name=True), plugin.type()))
            continue

        _init_builder(plugin, clazz_map[plugin.type()])


def setup_syntax_completion():
    try:
        import readline
    except ImportError:
        pass
    else:
        import rlcompleter
        readline.parse_and_bind("tab: complete")


def setup_excepthook(logger):
    """
    Setup the excepthook which logs uncaught exceptions instead of throwing them around.
    @see https://fman.io/blog/pyqt-excepthook/
    """
    def _excepthook(exc_type, exc_value, exc_tb):
        enriched_tb = _add_missing_frames(exc_tb) if exc_tb else exc_tb
        logger.debug("Uncaught exception", exc_info=(exc_type, exc_value, enriched_tb))

    def _add_missing_frames(tb):
        result = fake_tb(tb.tb_frame, tb.tb_lasti, tb.tb_lineno, tb.tb_next)
        frame = tb.tb_frame.f_back
        while frame:
            result = fake_tb(frame, frame.f_lasti, frame.f_lineno, result)
            frame = frame.f_back
        return result

    fake_tb = namedtuple(
        'fake_tb', ('tb_frame', 'tb_lasti', 'tb_lineno', 'tb_next')
    )

    sys.excepthook = _excepthook


def get_input(context, args):
    if args.file:
        try:
            with open(args.file, "r") as f:
                return f.read()
        except:
            context.logger().error("Error loading {file}. Aborting ...".format(file=args.file))
            sys.exit(1)
    if args.input:
        return args.input


def get_action_type(context, builder, name):
    return getattr(builder, name)


def get_plugin_action(context, action_type_name, action_type_method, method_name):
    try:
        return getattr(action_type_method(), method_name)
    except Exception as e:
        plugin_names = [name[:name.rindex('_')] for name in context.plugins().names(safe_names=True)]
        suggestions = [result for result in process.extract(method_name, plugin_names)]
        suggestion = 'Did you mean "{}"?'.format(suggestions[0][0]) if suggestions else ""
        context.logger().error(
            'No {type} named "{name}". {suggestion}'.format(
                type=action_type_name, name=method_name, suggestion=suggestion))
        sys.exit(1)


def get_plugin_config(context, arguments):
    result = {}
    for argument in arguments:
        sep_index = argument.find('=')
        if sep_index < 1:
            context.logger().error('Invalid argument specification! Expected key=value, got {}'.format(argument))
            sys.exit(1)

        # Parse key=value into name and value
        name = argument[:sep_index]
        value = argument[sep_index+1:]

        # Handle boolean values
        if value == "True" or value == "False":
            value = value == "True"

        result[name] = value
    return result

def main():
    # Abort program execution on ctrl+c
    signal.signal(signal.SIGINT, signal.SIG_DFL)

    # Loads logger, config and plugins.
    app_path = os.path.dirname(os.path.abspath(__file__))
    context = Context("net.bytebutcher.decoder_plus_plus", app_path, namespace=locals())

    try:

        # Builders can be used in interactive shell or within the ui's code-view.
        init_builder(context)

        parser = argparse.ArgumentParser(add_help=False)
        parser.add_argument('-?', '--help', action='store_true',
                            help="show this help message and exit")
        parser.add_argument('input', nargs='?',
                            help="specifies the input-text")
        parser.add_argument('-f', '--file', action=SingleArgs,
                            help="specifies the input-file")
        parser.add_argument('--new-instance', action='store_true',
                            help="opens new instance instead of new tab in already running instance.")
        parser.add_argument('--dialog', action='store_true',
                            help="opens a dialog which returns the transformed text when done.")
        parser.add_argument('-l', '--list-codecs', nargs='*', metavar="FILTER_TERM",
                            help="lists all available codecs or those matching the filter terms.")
        parser.add_argument('-e', '--encode', dest="encode", action=OrderedMultiArgs,
                            help="encodes the input using the specified codec(s).")
        parser.add_argument('-d', '--decode', action=OrderedMultiArgs,
                            help="decodes the input using the specified codec(s)")
        parser.add_argument('-h', '--hash', action=OrderedMultiArgs,
                            help="transforms the input using the specified hash-functions")
        parser.add_argument('-s', '--script', nargs='+', action=OrderedMultiArgs, metavar="OPTION=VALUE",
                            help="transforms the input using the specified script (optional arguments)")
        parser.add_argument('--debug', action='store_true',
                            help="activates debug mode with extensive logging. Output will be written into dpp.log "
                                 "inside the application directory.")

        args = parser.parse_args()
        if args.help:
            parser.print_help()
            sys.exit(0)

        if args.script and not args.file and not args.input:
            scripts = [arg[1] for arg in args.ordered_args if arg[0] == 'script']
            args.input = scripts.pop(-1).pop(-1)

        if args.debug:
            # Enable debug mode for current session.
            context.setDebugMode(True, temporary=True)

        # Start GUI when no other parameters were used.
        if not args.encode and not args.decode and not args.script and not args.hash and not type(args.list_codecs) == list:
            # Setup excepthook to handle uncaught exceptions.
            setup_excepthook(context.logger())
            # Update application mode
            context.setMode(Context.Mode.GRAPHICAL_UI)
            try:
                app = QApplication(sys.argv)
                instance_handler = InstanceHandler(app, context.getAppID())
                input = get_input(context, args)
                if args.dialog:
                    context.logger().info("Starting Decoder++ Dialog...")
                    ex = DecoderPlusPlusDialog(context, input)
                    sys.exit(app.exec_())
                if instance_handler.isAlreadyRunning():
                    context.logger().info("Application is already running...")
                    if args.new_instance:
                        context.logger().info("Starting Decoder++ GUI in new instance...")
                        ex = DecoderPlusPlusWindow(context, input)
                        sys.exit(app.exec_())
                    else:
                        context.logger().info("Opening new tab in already running instance...")
                        instance_handler.newTab(input)
                        sys.exit(0)
                else:
                    context.logger().info("Starting Decoder++ GUI...")
                    ex = DecoderPlusPlusWindow(context, input)
                    instance_handler.received.connect(ex.newTab)
                    sys.exit(app.exec_())
            except Exception as e:
                context.logger().exception("Unexpected Exception: {}".format(e), exc_info=context.isDebugModeEnabled())
                sys.exit(1)

        if type(args.list_codecs) == list:
            filter_terms = args.list_codecs
            row_format = "{:<20}  {}"
            print()
            print(row_format.format("Codec", "Type"))
            print(row_format.format("-----", "----"))
            for plugin in context.plugins():
                plugin_name = plugin.method_name().lower()
                plugin_type = plugin.type().lower()
                if filter_terms:
                    has_match = True
                    for filter_term in filter_terms:
                        filter_term = filter_term.lower()
                        if filter_term not in plugin_name and filter_term not in plugin_type:
                            has_match = False
                            break
                    if not has_match:
                        continue
                print(row_format.format(plugin_name, plugin_type))
            print()
            sys.exit(0)

        if not args.encode and not args.decode and not args.script and not args.hash:
            context.logger().error("No action specified!")
            sys.exit(1)

        if not args.file and not args.input:
            context.logger().error("No input specified!")
            sys.exit(1)

        if args.file and args.input:
            context.logger().error("Argument --file and input can not be used together.")
            sys.exit(1)

        # Command line usage
        context.setMode(Context.Mode.COMMAND_LINE)
        input = get_input(context, args)
        builder = DecoderPlusPlus(input)
        for name, values in args.ordered_args:
            if not values:
                # No input supplied (e.g. dpp --script search_and_replace)
                parser.print_help()
                sys.exit(1)
            method_name = values.pop(0)
            action_type = get_action_type(context, builder, name)
            plugin_action = get_plugin_action(context, name, action_type, method_name)
            show_plugin_help = "help" in values
            plugin_config = get_plugin_config(context, filter(lambda x: x != "help", values))
            if show_plugin_help: plugin_config["help"] = True
            builder = plugin_action(**plugin_config)

        print(builder.run())
    except Exception as e:
        context.logger().exception(e, exc_info=context.isDebugModeEnabled())


if __name__ == '__main__':
    main()
